/* -*- Mode: javascript; tab-width: 20; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mozilla Calendar code.
 *
 * The Initial Developer of the Original Code is
 *  Michiel van Leeuwen <mvl@exedo.nl>
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/**
 * A thin wrapper that is a print formatter, and just calls the html (list) 
 * exporter
 */

function FGcalListPrinter() {
    this.wrappedJSObject = this;
}

FGcalListPrinter.prototype.QueryInterface =
function QueryInterface(aIID) {
    if (!aIID.equals(Components.interfaces.nsISupports) &&
        !aIID.equals(Components.interfaces.calIPrintFormatter)) {
        throw Components.results.NS_ERROR_NO_INTERFACE;
    }

    return this;
};

FGcalListPrinter.prototype.getName =
function list_getName() {
    var sbs = Components.classes["@mozilla.org/intl/stringbundle;1"]
                    .getService(Components.interfaces.nsIStringBundleService);
    var props = sbs.createBundle("chrome://fgprinters/locale/fgprinters.properties");
    var retName = "FG " + props.GetStringFromName("listview");
    return retName;
};
FGcalListPrinter.prototype.__defineGetter__("name", FGcalListPrinter.prototype.getName);

FGcalListPrinter.prototype.formatToHtml =
function FGlistPrint_format(aStream, aStart, aEnd, aCount, aItems, aTitle) {

    // Create the e4x framework of the HTML document
    var html = <html/>;
    html.appendChild(
            <head>
                <title>{aTitle}</title>
                <meta http-equiv='Content-Type' content='text/html; charset=UTF-8'/>
                <style type='text/css'/>
            </head>);

    var fsizeday = getPrefSafe("fgprinters.list.fontsizedayname", 20);
    var dnbk = getPrefSafe("fgprinters.headcolour", "#f0f0f0");
    var oomc = getPrefSafe("fgprinters.othermonthcolour", "#EEEEFF");
    var wec = getPrefSafe("fgprinters.weekendcolour", "#FFFFEE");
    var sansserif = getPrefSafe("fgprinters.sansserif", true);
    var printLocation = getPrefSafe("fgprinters.list.printlocation", true);
    var printDescription = getPrefSafe("fgprinters.list.printdescription", true); 
    var printCategory = getPrefSafe("fgprinters.list.printcategory", true);
    var printHeader = getPrefSafe("fgprinters.list.printheader", true);
    var fsizeh = getPrefSafe("fgprinters.list.fontsizeheader", 14);
    var printStatus = getPrefSafe("fgprinters.list.printstatus", true);
    var printCalname = getPrefSafe("fgprinters.list.printcalname", false);
    var printtitle = getPrefSafe("fgprinters.printtitle", false);
    var fsizeglobaltitle = getPrefSafe("fgprinters.fontsizetitle", 26);
    var fsizesubtitle = getPrefSafe("fgprinters.fontsizesubtitle", 20);
    var subTitle = getPrefSafe("fgprinters.subtitle", "");
    var printMonths = getPrefSafe("fgprinters.list.printmonths", false);

    html.head.style += ".main-title { font-size: "+fsizeglobaltitle+"px; font-weight:bold; text-align: center; }\n";
    html.head.style += ".sub-title { font-size: "+fsizesubtitle+"px; font-weight:bold; text-align: center; }\n";
    html.head.style += ".day-name { border: 1px solid #000; background-color: "+dnbk+"; font-size: "+fsizeday+"px; font-weight: bold; }\n";
    html.head.style += ".day-box { border: 1px solid black; vertical-align: top; }\n";
    html.head.style += ".day-off { background-color: "+wec+" !important; }\n";
    html.head.style += ".out-of-month { background-color: "+oomc+" !important; }\n";
    if (sansserif) { html.head.style += "body, tr, td, table, p { font-family: Arial, sans-serif; }\n"; }
    html.head.style += ".weekend { background-color: "+wec+"; }\n";
    html.head.style += ".normal { }\n";
    
    var body = <body/>;

    if (printtitle && aTitle) {
        body.appendChild(<p class="main-title">{aTitle}</p>);
    }

    if (subTitle) {
        body.appendChild(<p class="sub-title">{subTitle}</p>);
    }


    // helper: returns the passed item's startDate, entryDate or dueDate, in
    //         that order. If the item doesn't have one of those dates, this
    //         doesn't return.
    function hasUsableDate(item) {
        return item.startDate || item.entryDate || item.dueDate;
    }

    // Clean out the item list so it only contains items we will want to
    // include in the printout.
    var filteredItems = aItems.filter(hasUsableDate);

    var calIEvent = Components.interfaces.calIEvent;
    var calITodo = Components.interfaces.calITodo
    function compareItems(a, b) {
        var start_a = calGetStartDate(a);
        if (!start_a) {
            return -1;
        }
        var start_b = calGetStartDate(b);
        if (!start_b) {
            return 1;
        }
        return start_a.compare(start_b);
    }
    var sortedList = filteredItems.sort(compareItems);

//    var weekFormatter = Components.classes["@mozilla.org/calendar/weektitle-service;1"]
//                                  .getService(Components.interfaces.calIWeekTitleService);

    // A place to cache the string for "All Day" events later if we need it.
    var mAllDayString = null;
    
    // Start at the beginning of the week that aStart is in, and loop until
    // we're at aEnd. In the loop we build the HTML table for each day, and
    // get the day's items using getDay().
    var start = aStart || sortedList[0].startDate || sortedList[0].entryDate ||
                sortList[0].dueDate;
    ASSERT(start, "can't find a good starting date to print");

    var lastItem = sortedList[sortedList.length-1];
    var end = aEnd || lastItem.startDate || lastItem.entryDate ||
               lastItem.dueDate;
    ASSERT(end, "can't find a good ending date to print");

    var date = start.clone();
    if (!date.isDate) {
        date.isDate = 1;
    } 
    
    var endDate = end.clone();
    if (!endDate.isDate) {
        endDate.isDate = 1;
        endDate.day += 1;
    }

    if (printMonths) {
         var prevMonth = date.clone();
         prevMonth.month--;
         prevMonth.day = 15;
         try {
             prevMonth.normalize();
         } catch(ex) {
         }
         var thisMonth = date.clone();
         var nextMonth = date.clone();
         nextMonth.month++;
         nextMonth.day = 15;
         try {
             nextMonth.normalize();
         } catch(ex) {
         }
         var prevMonthTable = this.createMonthTable(prevMonth);
         var thisMonthTable = this.createMonthTable(thisMonth);
         var nextMonthTable = this.createMonthTable(nextMonth);
         body.appendChild(
             <table border='0' width='100%' class='main-table'>
                    <tr>
                        <td width='1%' align='left' valign='center'>{prevMonthTable}</td>
                        <td width='49%' align='center' valign='center'><div style='color: transparent;'>a</div></td>
                        <td width='1%' align='center' valign='center'>{thisMonthTable}</td>
                        <td width='48%' align='center' valign='center'><div style='color: transparent;'>a</div></td>
                        <td width='1%' align='right' valign='center'>{nextMonthTable}</td>
                    </tr>
                </table>);
        } else {

    
    }

    const weekPrefix = "calendar.week.";
    var prefNames = ["d0sundaysoff", "d1mondaysoff", "d2tuesdaysoff",
                     "d3wednesdaysoff", "d4thursdaysoff", "d5fridaysoff", "d6saturdaysoff"];
    var defaults = [true, false, false, false, false, false, true];

    var mainTable = <table width='100%' border='1' rules='rows' cellpadding='5'/>;

    var dateFormatter = Components.classes["@mozilla.org/calendar/datetime-formatter;1"]
                                  .getService(Components.interfaces.calIDateTimeFormatter);

    if (printHeader)
    {
        var sbs = Components.classes["@mozilla.org/intl/stringbundle;1"]
	                    .getService(Components.interfaces.nsIStringBundleService);
        var props = sbs.createBundle("chrome://fgprinters/locale/fgprinters.properties");
		var cat = props.GetStringFromName("category");
		var stEn = props.GetStringFromName("startEnd");
		var event = props.GetStringFromName("event");
		var descr = props.GetStringFromName("description");
		var loc = props.GetStringFromName("location");
		var stat = props.GetStringFromName("status");
		var can = props.GetStringFromName("cal");
		if (!printCategory) {
			cat = " ";
		}
		if (!printLocation) {
			loc = " ";
		}
		if (!printDescription) {
			descr = " ";
		}
		if (!printStatus) {
			stat = " ";
		}
 		if (!printCalname) {
 			can = " ";
 		}
		var style = 'font-size: '+fsizeh+'px; color: black;';
		mainTable.appendChild(<tr style={style}><td> </td><td>{cat}</td><td>{stat}</td><td>{stEn}</td><td>{event}</td><td>{descr}</td><td>{loc}</td><td>{can}</td></tr>);
    }
    try {
		var defaultTimezone = calendarDefaultTimezone(); 
	} catch(ex) {
	}
    
	while(date.compare(endDate) == -1) {
		var isOff = getPrefSafe(weekPrefix+prefNames[date.weekday], defaults[date.weekday]);
		var oneDay = this.getDay(date, sortedList, isOff);
		if (oneDay) {
			try {
				var dateString = dateFormatter.formatDateLong(date.getInTimezone(defaultTimezone));
			} catch(ex) {
				var dateString = dateFormatter.formatDateLong(date);
			}
			mainTable.appendChild(<tr style='page-break-after: avoid;'><td colspan='13' align='center' class='day-name'>{dateString}</td></tr>);
			mainTable.appendChild(oneDay);
		}
		date.day += 1;
		try {
			date.normalize();
		} catch(ex) {
		}
    }
    
    body.appendChild(mainTable);
    
    html.appendChild(body);

    // Stream out the resulting HTML
    var convStream = Components.classes["@mozilla.org/intl/converter-output-stream;1"]
                               .getService(Components.interfaces.nsIConverterOutputStream);
    convStream.init(aStream, 'UTF-8', 0, 0x0000);
    convStream.writeString(html.toXMLString());
};

FGcalListPrinter.prototype.getDay =
/**
 * Given a calIDateTime and an array of items, this function creates an HTML
 * table containing the items, using the appropriate formatting and colours.
 */
function FGlistPrint_getDayTable(aDate, aItems, isOff) {
    // mainTd is the <td> element from the parent HTML table that will hold
    // the child HTML tables containing the date string and this day's items.

	var nofEvents = 0;
    
    var mainTd = <div/>;
    
    var dateFormatter = Components.classes["@mozilla.org/calendar/datetime-formatter;1"]
                                  .getService(Components.interfaces.calIDateTimeFormatter);
	
    var printAllDay = getPrefSafe("fgprinters.list.printallday", false);
    var printLocation = getPrefSafe("fgprinters.list.printlocation", true);
    var printDescription = getPrefSafe("fgprinters.list.printdescription", true); 
    var printCategory = getPrefSafe("fgprinters.list.printcategory", true);
    var printCalColor = getPrefSafe("fgprinters.list.printcalcolor", true);
    var printStatus = getPrefSafe("fgprinters.list.printstatus", true);
    var printPriority = getPrefSafe("fgprinters.list.printpriority", true);
	var printCalname = getPrefSafe("fgprinters.list.printcalname", false);
	
    var fsize = getPrefSafe("fgprinters.list.fontsize", 14);
    var obeyPrivacy = getPrefSafe("fgprinters.obeyprivacy", false);
    try {
		var defaultTimezone = calendarDefaultTimezone(); 
	} catch(ex) {
	    var defaultTimezone = null;
	}
        
    // Add the formatted items (in their child HTML table)
        
    for each (var item in aItems) {
        var sDate = item.startDate || item.entryDate || item.dueDate;
        var eDate = item.endDate || item.dueDate || item.entryDate;
        if (defaultTimezone && sDate) {
			try {
				sDate = sDate.getInTimezone(defaultTimezone);
			} catch(ex) {
			}
        }
        if (defaultTimezone && eDate) {
            try {
				eDate = eDate.getInTimezone(defaultTimezone);
			} catch(ex) {
			}
        }
        
        // End dates are exclusive. Adjust the eDate accordingly.
        if (sDate && sDate.isDate && eDate) {
            eDate = eDate.clone();
            eDate.day -= 1;
            try {
                eDate.normalize();
            } catch(ex) {
            }
        }

        // If the item has no end date, or if the item's end date is aDate or
        // is before aDate, skip to the next item.
        if (!eDate || (eDate.compare(aDate) < 0)) {
            continue;
        }
       
        // No start date or a start date that's after the date we want is bad.
        if (!sDate || (sDate.compare(aDate) > 0)) {
            break;
        }

        function getStringForDate(oneDate) {
            if (!oneDate.isDate) {
                return dateFormatter.formatTime(oneDate);
            }

            // We cache the string for "All Day" 'cause we're good like that.
            if (this.mAllDayString == null) {
                var sbs = Components.classes["@mozilla.org/intl/stringbundle;1"]
                                    .getService(Components.interfaces.nsIStringBundleService);
                var props = sbs.createBundle("chrome://calendar/locale/dateFormat.properties");
                this.mAllDayString = props.GetStringFromName("AllDay");
            }
            if (printAllDay) {
				return this.mAllDayString;
			} else { 
				return "";
			}
        }

		nofEvents++;

        var time;
        if (sDate && eDate) {
            if (!sDate.isDate) {
				var startTime = "";
				var endTime = "";
				if (sDate.compare(aDate) < 0) {
					var dummyDate = sDate.clone();
					dummyDate.hour = 0;
					dummyDate.minute = 0;
					try {
						dummyDate.normalize();
					} catch(ex) {
					}
					startTime = getStringForDate(dummyDate);
				} else {
					startTime = getStringForDate(sDate);
				}
				if (eDate.compare(aDate) > 0) {
					var dummyDate = eDate.clone();
					dummyDate.hour = 23;
					dummyDate.minute = 59;
					try {
						dummyDate.normalize();
					} catch(ex) {
					}
					endTime = getStringForDate(dummyDate);
				} else {
					endTime = getStringForDate(eDate);
				}
                time = startTime + '-' + endTime;
            } else {
                time = getStringForDate(sDate);
            }
        } else if (sDate) {
            time = getStringForDate(sDate);
        } else if (eDate) {
            time = getStringForDate(eDate);
        }

        var cat = item.getProperty("CATEGORIES");
        
        if (!cat) {
            cat = " ";
            for each (var catT in item.getCategories({})) {
                try {
                    cat = catT;
                    break; // take first matching
                } catch(ex) {
                    cat = " ";
                }
            }
        }
        if (!printCategory) {
            cat = " ";
        }

		var style = 'font-size: '+fsize+'px; page-break-inside: avoid; color: black;';

		if (printCalColor) {
            try {
				var calColor = item.calendar.getProperty('color');
			} catch(ex) {
				var calMgr = Components.classes["@mozilla.org/calendar/manager;1"]
									.getService(Components.interfaces.calICalendarManager);
				var calColor = calMgr.getCalendarPref(item.calendar, 'color'); 
			}
			style += 'border: solid ' + calColor + ' 2px;';
		}
        var location = " ";
        var descr = " ";
		var can = " ";

        if (printLocation)
        {
            if (item.hasProperty('LOCATION')) {
                location = item.getProperty('LOCATION');
            }
        }
        if (printDescription)
        {
            if (item.hasProperty('DESCRIPTION')) {
                descr = item.getProperty('DESCRIPTION');
            }
        }
		if (printCalname)
		{
			can = item.calendar.getProperty('name');
		}

        var title = item.title;
        if (obeyPrivacy && item.privacy != "PUBLIC") {
            title = " ";
            descr = " ";
            location = " ";
        }
        var trClass = "normal";
        if (isOff) {
			trClass = "weekend";
		}
		var status = " ";
		if (item.status && printStatus) {
			if (item.status == "CONFIRMED") {
				status = calGetString("calendar", "statusConfirmed");
			}
			if (item.status == "TENTATIVE") {
				status = calGetString("calendar", "statusTentative");
			}
			if (item.status == "CANCELLED") {
				status = calGetString("calendar", "statusCancelled");
			}
		}
		var prioCol = "transparent";
		if (item.priority && printPriority) {
			if (item.priority > 5) {
				prioCol = "#0F0";
			}
			if (item.priority == 5) {
				prioCol = "#FF0";
			}
			if (item.priority < 5) {
				prioCol = "#F00";
			}
		}
		var prioMarker = 'background-color: ' + prioCol + ';';
        var item = <tr width='100%' class={trClass} style={style}>
                       <td style={prioMarker}> </td><td>{cat}</td><td>{status}</td><td>{time}</td><td>{title}</td><td>{descr}</td><td>{location}</td><td>{can}</td>
                   </tr>;
		mainTd.appendChild(item);
		if (printCalColor) {
			mainTd.appendChild(<div style='margin-top: 2px'></div>);
		}
    }
    if (nofEvents) {
		return mainTd;
	} else {
		return;
	}
};

FGcalListPrinter.prototype.createMonthTable =
function FGlistPrint_doOtherMonth(aMonth) {
    var weekStart = getPrefSafe("calendar.week.start", 0);

    var monthName = calGetString("dateFormat", "month." + (aMonth.month +1)+ ".name") + " " + aMonth.year;

    var fsize = getPrefSafe("fgprinters.list.fontsizeom", 10);
    var tStyle = 'font-size: ' + fsize + 'px;';

    var monthTable = <table width='100%' rules='both' style={tStyle}><tr><td colspan='7' align='center'>{monthName}</td></tr></table>;
    var firstDate = aMonth.startOfMonth.startOfWeek.clone();
    firstDate.day += weekStart;
    try {
        firstDate.normalize();
    } catch(ex) {
    }
    if (aMonth.startOfMonth.weekday < weekStart) {
        // Go back one week to make sure we display this day
        firstDate.day -= 7;
        try {
            firstDate.normalize();
        } catch(ex) {
        }
    }

    var lastDate = aMonth.endOfMonth.endOfWeek.clone();
    if (aMonth.endOfMonth.weekday < weekStart) {
        // Go back one week so we don't display any extra days
        lastDate.day -= 7;
        try {
            lastDate.normalize();
        } catch(ex) {
        }
    }
    firstDate.isDate = true;
    lastDate.isDate = true;

    var printWeekDays = getPrefSafe("fgprinters.printweekdaysmm", false);
    if (printWeekDays) {
        const weekPrefix = "calendar.week.";
        var prefNames = ["d0sundaysoff", "d1mondaysoff", "d2tuesdaysoff",
                         "d3wednesdaysoff", "d4thursdaysoff", "d5fridaysoff", "d6saturdaysoff"];
        var defaults = [true, false, false, false, false, false, true];
        var daysOff = new Array();
        for (var i in prefNames) {
            if (getPrefSafe(weekPrefix+prefNames[i], defaults[i])) {
                daysOff.push(Number(i));
            }
        }
        var weekRow = <tr/>;
        for (var i = 0; i < 7; i++) {
            var myClass = 'day-box';
            var style = 'padding-left: 0.2em; padding-right: 0.2em; ';
            var dayName = calGetString("dateFormat", "day."+ (((i + weekStart) % 7)+1) + ".Mmm")[0];
            var day = <td align='center' valign='center' class={myClass}><div style={style}>{dayName}</div></td>
            weekRow.appendChild(day);
            try {
                date.normalize();
            } catch(ex) {
            }
        }
        monthTable.appendChild(weekRow);
    }
    
    var date = firstDate.clone();
    var itemListIndex = 0;
    while (date.compare(lastDate) != 1) {
        monthTable.appendChild(this.makeOtherMonthWeek(date, aMonth.month));
    }
    return monthTable;
};

FGcalListPrinter.prototype.makeOtherMonthWeek =
function FGmakeOtherMonthWeek(date, targetMonth) {
    var weekRow = <tr/>;
    var printOMDays = getPrefSafe("fgprinters.printomdaysmm", true);
    const weekPrefix = "calendar.week.";
    var prefNames = ["d0sundaysoff", "d1mondaysoff", "d2tuesdaysoff",
                     "d3wednesdaysoff", "d4thursdaysoff", "d5fridaysoff", "d6saturdaysoff"];
    var defaults = [true, false, false, false, false, false, true];
    var daysOff = new Array();
    for (var i in prefNames) {
        if (getPrefSafe(weekPrefix+prefNames[i], defaults[i])) {
            daysOff.push(Number(i));
        }
    }

    for (var i = 0; i < 7; i++) {
        var myClass = 'day-box';
        var style = 'padding-left: 0.2em; padding-right: 0.2em;';
        if (date.month != targetMonth) {
            myClass += ' out-of-month';
            if (!printOMDays) {
                style += 'color: transparent; ';
            }
        } else if (daysOff.some(function(a) { return a == date.weekday; })) {
            myClass += ' day-off';
        }
        var day = <td align='right' valign='center' class={myClass}><div style={style}>{date.day}</div></td>
        weekRow.appendChild(day);
        date.day++;
        try {
            date.normalize();
        } catch(ex) {
        }
    }
    return weekRow;
};



/**
 * Normal get*Pref calls will throw if the pref is undefined.  This function
 * will get a bool, int, or string pref.  If the pref is undefined, it will
 * return aDefault.
 *
 * @param aPrefName   the (full) name of preference to get
 * @param aDefault    (optional) the value to return if the pref is undefined
 */
function getPrefSafe(aPrefName, aDefault) {
    const nsIPrefBranch = Components.interfaces.nsIPrefBranch;
    const prefB = Components.classes["@mozilla.org/preferences-service;1"]
                            .getService(nsIPrefBranch);
    switch (prefB.getPrefType(aPrefName)) {
        case nsIPrefBranch.PREF_BOOL:
            return prefB.getBoolPref(aPrefName);
        case nsIPrefBranch.PREF_INT:
            return prefB.getIntPref(aPrefName);
        case nsIPrefBranch.PREF_STRING:
            return prefB.getCharPref(aPrefName);
        default: // includes nsIPrefBranch.PREF_INVALID
            return aDefault;
    }
}

/**
 * Logs a message and the current js-stack, if aCondition fails
 *
 * @param aCondition  the condition to test for
 * @param aMessage    the message to report in the case the assert fails
 * @param aCritical   if true, throw an error to stop current code execution
 *                    if false, code flow will continue
 */
function ASSERT(aCondition, aMessage, aCritical) {
    if (aCondition) {
        return;
    }

    var string = "Assert failed: " + aMessage + '\n' + STACK();
    if (aCritical) {
        throw new Error(string);
    } else {
        Components.utils.reportError(string);
    }
}

function calGetStartDate(aItem)
{
    return (isEvent(aItem) ? aItem.startDate : aItem.entryDate);
}

function isEvent(aObject) {
    return aObject instanceof Components.interfaces.calIEvent;
} 

/**
 * Function to get the (cached) best guess at a user's default timezone.  We'll
 * use the value of the calendar.timezone.local preference, if it exists.  If
 * not, we'll do our best guess.
 *
 * @return user's default timezone.
 */
function calendarDefaultTimezone() {
    if (calendarDefaultTimezone.mTz === undefined) {
        var prefTzid = getPrefSafe("calendar.timezone.local", null);
        var tzid = prefTzid;
        if (!tzid) {
            tzid = guessSystemTimezone();
        }
        calendarDefaultTimezone.mTz = getTimezoneService().getTimezone(tzid);
        ASSERT(calendarDefaultTimezone.mTz, "timezone not found: " + tzid);
        // Update prefs if necessary:
        if (calendarDefaultTimezone.mTz && calendarDefaultTimezone.mTz.tzid != prefTzid) {
            setPref("calendar.timezone.local", "CHAR", calendarDefaultTimezone.mTz.tzid);
        }
    }
    return calendarDefaultTimezone.mTz;
}

function getTimezoneService() {
    if (getTimezoneService.mObject === undefined) {
        getTimezoneService.mObject = Components.classes["@mozilla.org/calendar/timezone-service;1"]
                                               .getService(Components.interfaces.calITimezoneService);
    }
    return getTimezoneService.mObject;
}
