/* -*- Mode: javascript; tab-width: 20; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mozilla Calendar code.
 *
 * The Initial Developer of the Original Code is
 *   Joey Minta <jminta@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Matthew Willis <mattwillis@gmail.com>
 *   Ferdinand Grassmann <ferdinand@grassmann.info>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/**
 * Prints a rough month-grid of events/tasks
 */

function FGcalMonthPrinter() {
    this.wrappedJSObject = this;
}

FGcalMonthPrinter.prototype.QueryInterface =
function QueryInterface(aIID) {
    if (!aIID.equals(Components.interfaces.nsISupports) &&
        !aIID.equals(Components.interfaces.calIPrintFormatter)) {
        throw Components.results.NS_ERROR_NO_INTERFACE;
    }

    return this;
};

FGcalMonthPrinter.prototype.getName =
function FGmonthPrint_getName() {
    return "FG "+calGetString("calendar", "monthPrinterName");    
};
FGcalMonthPrinter.prototype.__defineGetter__("name", FGcalMonthPrinter.prototype.getName);

FGcalMonthPrinter.prototype.formatToHtml =
function FGmonthPrint_format(aStream, aStart, aEnd, aCount, aItems, aTitle) {

    var html = <html/>
    html.appendChild(
            <head>
                <title>{aTitle}</title>
                <meta http-equiv='Content-Type' content='text/html; charset=UTF-8'/>
                <style type='text/css'/>
            </head>);
    var fontsizetitle = getPrefSafe("fgprinters.monthly.fontsizetitle", 26);
    var fontsizeday = getPrefSafe("fgprinters.monthly.fontsizeday", 12);
    var dnbk = getPrefSafe("fgprinters.headcolour", "#f0f0f0");
    var oomc = getPrefSafe("fgprinters.othermonthcolour", "#EEEEFF");
    var wec = getPrefSafe("fgprinters.weekendcolour", "#FFFFEE");
    var sansserif = getPrefSafe("fgprinters.sansserif", true);
    var printOM = getPrefSafe("fgprinters.monthly.printothermonths", false);

    var fsizeloc = getPrefSafe("fgprinters.monthly.fontsizelocation", 8);
    var fsizedesc = getPrefSafe("fgprinters.monthly.fontsizedescription", 8);
    var printLocationRight = getPrefSafe("fgprinters.monthly.printlocationright", true);
    var printtitle = getPrefSafe("fgprinters.printtitle", false);
    var fsizeglobaltitle = getPrefSafe("fgprinters.fontsizetitle", 26);
    var fsizesubtitle = getPrefSafe("fgprinters.fontsizesubtitle", 20);
    var fsizeweekno = getPrefSafe("fgprinters.monthly.fontsizeweekno", 12);
    var subTitle = getPrefSafe("fgprinters.subtitle", "");

    html.head.style = ".main-table { font-size: "+fontsizetitle+"px; font-weight: bold; }\n";
    html.head.style += ".sub-title { font-size: "+fsizesubtitle+"px; font-weight:bold; text-align: center; }\n";
    html.head.style += ".main-title { font-size: "+fsizeglobaltitle+"px; font-weight:bold; text-align: center; }\n";
    html.head.style += ".day-name { border: 1px solid black; background-color: "+dnbk+"; font-size: "+fontsizeday+"px; font-weight: bold; }\n";
    html.head.style += ".day-box { border: 1px solid black; vertical-align: top; }\n";
    html.head.style += ".day-box-sun { border-top: 1px solid black; vertical-align: top; }\n";
    html.head.style += ".out-of-month { background-color: "+oomc+" !important; }\n";
    html.head.style += ".day-off { background-color: "+wec+" !important; }\n";
    html.head.style += ".weekno { font-size: "+fsizeweekno+"px; }\n";

    if (printLocationRight)
    {
        html.head.style += ".loc { font-size: "+fsizeloc+"px; text-align: right; }\n";
    } else {
        html.head.style += ".loc { font-size: "+fsizeloc+"px; text-align: left; }\n";
    }
    html.head.style += ".desc { font-size: "+fsizedesc+"px; text-align: left; }\n";

    if (sansserif) { html.head.style += "body, table, tr, td, p { font-family: Arial, sans-serif; }\n"; }

    // If aStart or aEnd weren't passed in, we need to calculate them based on
    // aItems data.

    var start = aStart;
    var end = aEnd;
    if (!start || !end) {
        for each (var item in aItems) {
            var itemStart = item.startDate || item.entryDate;
            var itemEnd = item.endDate || item.dueDate;
            if (!start || (itemStart && start.compare(itemStart) == 1)) {
                start = itemStart;
            }
            if (!end || (itemEnd && end.compare(itemEnd) == -1)) {
                end = itemEnd;
            }
        }
    }
    // Play around with aStart and aEnd to determine the minimal number of
    // months we can show to still technically meet their requirements.  This
    // is most useful when someone printed 'Current View' in the month view. If
    // we take the aStart and aEnd literally, we'll print 3 months (because of
    // the extra days at the start/end), but we should avoid that.
    //
    // Basically, we check whether aStart falls in the same week as the start
    // of a month (ie aStart  is Jan 29, which often is in the same week as
    // Feb 1), and similarly whether aEnd falls in the same week as the end of
    // a month.
    var weekStart = getPrefSafe("calendar.week.start", 0);
    maybeNewStart = start.clone();
    maybeNewStart.day = 1;
    maybeNewStart.month = start.month+1;
    try {
        maybeNewStart.normalize();
    } catch(ex) {
    }

    var date = start.clone();

    // First we have to adjust the end date for comparison, as the
    // provided end date is exclusive, i.e. will not be displayed.

    var realEnd = end.clone();
    realEnd.day -= 1;
    try {
        realEnd.normalize();
    } catch(ex) {
    }

    if (start.compare(realEnd) <= 0) {
        // Only adjust dates if start date is earlier than end date.

        if ((start.month != realEnd.month) || (start.year != realEnd.year)) {
            // We only need to adjust if start and end are in different months.

            // We want to check whether or not the start day is in the same
            // week as the beginning of the next month. To do this, we take
            // the start date, add seven days and subtract the "day of week"
            // value (which has to be corrected in case we do not start on
            // Sunday).
            var testBegin = start.clone();
            var startWeekday = testBegin.weekday;
            if (startWeekday < weekStart) {
                startWeekday += 7;
            }
            testBegin.day += 7 + weekStart - startWeekday;
            try {
                testBegin.normalize();
            } catch(ex) {
            }
            if (testBegin.compare(maybeNewStart) > 0) {
                start = maybeNewStart;
                date = start.clone();
            }
        }
        if ((start.month != realEnd.month) || (start.year != realEnd.year)) {
            // We only need to adjust if start and end are in different months.

            // Next, we want to check whether or not the end day is in the same
            // week as the end of the previous month. So we have to get the
            // "day of week" value for the end of the previous month, adjust it
            // if necessary (when start of week is not Sunday) and check if the
            // end day is in the same week.

            var lastDayOfPreviousMonth = end.clone();
            lastDayOfPreviousMonth.day = 0;
            try {
                lastDayOfPreviousMonth.normalize();
            } catch(ex) {
            }
            var lastDayWeekday = lastDayOfPreviousMonth.weekday;
            if (lastDayWeekday < weekStart) {
                lastDayWeekday += 7;
            }
            if (date.month != end.month) {
                date.day = 1;
            }
            if ((lastDayWeekday + end.day - 1) < (7 + weekStart)) {
                date.day = end.day;
            }
            try {
                date.normalize();
            } catch(ex) {
            }

            // Finally, we have to check whether we adjusted the dates too
            // well so that nothing is printed. That happens if you print just
            // one week which has the last day of a month in it.

            if (date.compare(end) >= 0) {
                date.day = 1;
                try {
                    date.normalize();
                } catch(ex) {
                }
            }
        } else {
            date.day = 1;
            try {
                date.normalize();
            } catch(ex) {
            }
        }
    } else {
         // If start date is after end date, just print empty month.
         date = realEnd.clone();
    }

    var body = <body/>

    if (printtitle && aTitle) {
        body.appendChild(<p class="main-title">{aTitle}</p>);
    }

    if (subTitle) {
        body.appendChild(<p class="sub-title">{subTitle}</p>);
    }

    while (date.compare(end) < 0) {
        var monthName = calGetString("dateFormat", "month." + (date.month +1)+ ".name");
        monthName += " " + date.year;

        if (printOM) {
            var prevMonth = date.clone();
            prevMonth.month--;
            prevMonth.day = 15;
            try {
                prevMonth.normalize();
            } catch(ex) {
            }
            var nextMonth = date.clone();
            nextMonth.month++;
            nextMonth.day = 15;
            try {
                nextMonth.normalize();
            } catch(ex) {
            }
            var prevMonthTable = this.createMonthTable(prevMonth);
            var nextMonthTable = this.createMonthTable(nextMonth);
            body.appendChild(
                <table border='0' width='100%' class='main-table'>
                    <tr>
                        <td width='1%' align='left' valign='center'>{prevMonthTable}</td>
                        <td width='98%' align='center' valign='center'>{monthName}</td>
                        <td width='1%' align='right' valign='center'>{nextMonthTable}</td>
                    </tr>
                </table>);
        } else {
            body.appendChild(
                <table border='0' width='100%' class='main-table'>
                    <tr>
                        <td align='center' valign='center'>{monthName}</td>
                    </tr>
                </table>);
        }
        body.appendChild(this.getStringForMonth(date, aItems));
        // Make sure each month gets put on its own page
        body.appendChild(<br style="page-break-after:always;"/>);
        date.month++;
        try {
            date.normalize();
        } catch(ex) {
        }
    }
    html.appendChild(body);
    
    var convStream = Components.classes["@mozilla.org/intl/converter-output-stream;1"]
                               .getService(Components.interfaces.nsIConverterOutputStream);
    convStream.init(aStream, 'UTF-8', 0, 0x0000);
    convStream.writeString(html.toXMLString());
//    Components.utils.reportError(html.toXMLString());
};

FGcalMonthPrinter.prototype.getStringForMonth =
function FGmonthPrint_getHTML(aStart, aItems) {
    var weekStart = getPrefSafe("calendar.week.start", 0);
    var smallSatSun = getPrefSafe("fgprinters.monthly.smallsatsun", false);
    var printWeekNo = getPrefSafe("fgprinters.monthly.printweekno", false);

    var monthTable = <table style='border:1px solid black;' width='100%' rules='all'/>
    var dayNameRow = <tr/>
    if (printWeekNo) {
        var sbs = Components.classes["@mozilla.org/intl/stringbundle;1"]
                            .getService(Components.interfaces.nsIStringBundleService);
        var props = sbs.createBundle("chrome://calendar/locale/calendar.properties");
        var weekName = " " + props.GetStringFromName('shortcalendarweek') + " "; 
        dayNameRow.appendChild(<td class='day-name' align='center' width='0%'>{weekName}</td>);
    }
    if (!smallSatSun) {
        for (var i = 0; i < 7; i++) {
            var dayName = calGetString("dateFormat", "day."+ (((weekStart+i)%7)+1) + ".Mmm");
            dayNameRow.appendChild(<td class='day-name' align='center' width='14%'>{dayName}</td>);
        }
    } else {
        for (var i = 0; i < 5; i++) {
            var dayName = calGetString("dateFormat", "day."+ (((weekStart+i)%7)+1) + ".Mmm");
            dayNameRow.appendChild(<td class='day-name' align='center' width='16%'>{dayName}</td>);
        }
        var dayName = calGetString("dateFormat", "day."+ (((weekStart+5)%7)+1) + ".Mmm") + '/' + calGetString("dateFormat", "day."+ (((weekStart+6)%7)+1) + ".Mmm");
        dayNameRow.appendChild(<td class='day-name' align='center' width='16%'>{dayName}</td>);
    }
    monthTable.appendChild(dayNameRow);

    // Set up the item-list so it's easy to work with.
    function hasUsableDate(item) {
        return item.startDate || item.entryDate || item.dueDate;
    }
    var filteredItems = aItems.filter(hasUsableDate);

    var calIEvent = Components.interfaces.calIEvent;
    var calITodo = Components.interfaces.calITodo

    try {
        var defaultTimezone = calendarDefaultTimezone(); 
    } catch(ex) {
        var defaultTimezone = null;
    }

    function compareItems(a, b) {
        // Sort tasks before events
        if (a instanceof calIEvent && b instanceof calITodo) {
            return 1;
        }
        if (a instanceof calITodo && b instanceof calIEvent) {
            return -1;
        }

        if (a instanceof calIEvent) {
            if (a.startDate.isDate && b.startDate.isDate)
            {
                return a.startDate.compare(b.startDate);
            }
            if (a.startDate.isDate)
            {
                var tmpd = b.startDate.clone();
                tmpd.isDate = true;
                try {
                    tmpd.normalize();
                } catch(ex) {
                }
                if (defaultTimezone) {
                    try {
                        tmpd = tmpd.getInTimezone(defaultTimezone);
                    } catch(ex) {
                    }
                }
                if (a.startDate.compare(tmpd) == 0) return -1;
            }
            if (b.startDate.isDate)
            {
                var tmpd = a.startDate.clone();
                tmpd.isDate = true;
                try {
                    tmpd.normalize();
                } catch(ex) {
                }
                if (defaultTimezone) {
                    try {
                        tmpd = tmpd.getInTimezone(defaultTimezone);
                    } catch(ex) {
                    }
                }
                if (b.startDate.compare(tmpd) == 0) return 1;
            }
        }
        
        if (a instanceof calIEvent) {
            var startCompare = a.startDate.compare(b.startDate);
            if (startCompare != 0) {
                return startCompare;
            }
            return a.endDate.compare(b.endDate);
        }
        var aDate = a.entryDate || a.dueDate;
        var bDate = b.entryDate || b.dueDate;
        return aDate.compare(bDate);
    }
    var sortedList = filteredItems.sort(compareItems);
    var firstDate = aStart.startOfMonth.startOfWeek.clone();
    firstDate.day += weekStart;
    try {
        firstDate.normalize();
    } catch(ex) {
    }
    if (aStart.startOfMonth.weekday < weekStart) {
        // Go back one week to make sure we display this day
        firstDate.day -= 7;
        try {
            firstDate.normalize();
        } catch(ex) {
        }
    }

    var lastDate = aStart.endOfMonth.endOfWeek.clone();
    if (aStart.endOfMonth.weekday < weekStart) {
        // Go back one week so we don't display any extra days
        lastDate.day -= 7;
        try {
            lastDate.normalize();
        } catch(ex) {
        }
    }
    firstDate.isDate = true;
    lastDate.isDate = true;

    var date = firstDate.clone();
    var itemListIndex = 0;
    while (date.compare(lastDate) != 1) {
        monthTable.appendChild(this.makeHTMLWeek(date, sortedList, aStart.month));
    }
    return monthTable;
};

function getContrastingTextColor(bgColor)
{
    var calcColor = bgColor.replace(/#/g, "");
    var red = parseInt(calcColor.substring(0, 2), 16);
    var green = parseInt(calcColor.substring(2, 4), 16);
    var blue = parseInt(calcColor.substring(4, 6), 16);

    // Calculate the L(ightness) value of the HSL color system.
    // L = (max(R, G, B) + min(R, G, B)) / 2
    var max = Math.max(Math.max(red, green), blue);
    var min = Math.min(Math.min(red, green), blue);
    var lightness = (max + min) / 2;

    // Consider all colors with less than 50% Lightness as dark colors
    // and use white as the foreground color; otherwise use black.
    // Actually we use a threshold a bit below 50%, so colors like
    // #FF0000, #00FF00 and #0000FF still get black text which looked
    // better when we tested this.
    if (lightness < 120) {
        return "white";
    }
    
    return "black";
}

FGcalMonthPrinter.prototype.makeHTMLDay =
function FGmakeHTMLDay(date, sortedList, targetMonth, smallDay) {


    var obeyPrivacy = getPrefSafe("fgprinters.obeyprivacy", false);
    const weekPrefix = "calendar.week.";
    var prefNames = ["d0sundaysoff", "d1mondaysoff", "d2tuesdaysoff",
                     "d3wednesdaysoff", "d4thursdaysoff", "d5fridaysoff", "d6saturdaysoff"];
    var defaults = [true, false, false, false, false, false, true];
    var daysOff = new Array();
    for (var i in prefNames) {
        if (getPrefSafe(weekPrefix+prefNames[i], defaults[i])) {
            daysOff.push(Number(i));
        }
    }

    var printAllDay = getPrefSafe("fgprinters.monthly.printallday", false);
    var printPriority = getPrefSafe("fgprinters.monthly.printpriority", true);
    var printStatus = getPrefSafe("fgprinters.monthly.printstatus", true);
    var fsize = getPrefSafe("fgprinters.monthly.fontsize", 10);
    var fsizet = getPrefSafe("fgprinters.monthly.fontsizetime", 10);
    var oldcolour = getPrefSafe("fgprinters.monthly.oldcolour", false);
    var printLocation = getPrefSafe("fgprinters.monthly.printlocation", false);
    var printDescription = getPrefSafe("fgprinters.monthly.printdescription", true);
    var smallSatSun = getPrefSafe("fgprinters.monthly.smallsatsun", false);
    var printDayLeft = getPrefSafe("fgprinters.monthly.aligndayleft", false);
    var noOtherMonth = getPrefSafe("fgprinters.monthly.noothermonth", false);
    var printEndTime = getPrefSafe("fgprinters.monthly.printendtime", true);
    var noPrintTimesInMultiDay = getPrefSafe("fgprinters.monthly.noprinttimesinmultidayevents", false);
    
    var noEventsToday = noOtherMonth && (date.month != targetMonth);

    try {
        var defaultTimezone = calendarDefaultTimezone(); 
    } catch(ex) {
        var defaultTimezone = null;
    }

    if (smallDay < 2) {
        var myClass = 'day-box';
    } else {
        var myClass = 'day-box-sun';
    }
    if (date.month != targetMonth) {
        myClass += ' out-of-month';
    } else if (daysOff.some(function(a) { return a == date.weekday; })) {
        myClass += ' day-off';
    }

    var height = '100%';

    if (smallDay) {
        height = '50%';
    }

    var dayTd = <td align='left' valign='top' class={myClass} height={height} width='100'/>

    var tStyle = 'font-size: ' + fsize + 'px;';

    if (smallDay < 2) {
        var innerTable = <table width='100%' valign='top' border='0' height={height}/>
    } else {
        var innerTable = <table width='100%' valign='top' border='0' height={height} class={myClass}/>
    }
    var dayNum = date.day;
    if (noEventsToday) {
      dayNum = " ";
    }
    if (printDayLeft) {
        var dateLabel = <tr valign='top' height='1px'>
                            <td valign='top' align='left' style={tStyle}>{dayNum}</td>
                        </tr>
    } else {
        var dateLabel = <tr valign='top' height='1px'>
                            <td valign='top' align='right' style={tStyle}>{dayNum}</td>
                        </tr>
    }
    innerTable.appendChild(dateLabel);
  if (!noEventsToday) {
    for each (var item in sortedList) {
        var sDate = item.startDate || item.entryDate || item.dueDate;
        var eDate = item.endDate || item.dueDate || item.entryDate;
        if (defaultTimezone && sDate) {
            try {
                sDate = sDate.getInTimezone(defaultTimezone);
            } catch(ex) {
            }
        }
        if (defaultTimezone && eDate) {
            try {
                eDate = eDate.getInTimezone(defaultTimezone);
            } catch(ex) {
            }
        }

            // end dates are exclusive
        if (sDate.isDate) {
            eDate = eDate.clone();
            eDate.day -= 1;
            try {
                eDate.normalize();
            } catch(ex) {
            }
        }
        if (!eDate || eDate.compare(date) == -1) {
            continue;
        }
        itemListIndex = i;
        if (!sDate || sDate.compare(date) == 1) {
            break;
        }
        var dateFormatter = 
            Components.classes["@mozilla.org/calendar/datetime-formatter;1"]
                      .getService(Components.interfaces.calIDateTimeFormatter);


        function getStringForDate(date) {
            var dstring;
            if (!date.isDate) {
                return dateFormatter.formatTime(date);
            }
            if (printAllDay) { return calGetString("dateFormat", "AllDay"); } else { return ""; }
        }

        var time;
        if (sDate && eDate) {
            if (!sDate.isDate) {
                var startTime = "";
                var endTime = "";
                if (sDate.compare(date) < 0) {
                    var dummyDate = sDate.clone();
                    dummyDate.hour = 0;
                    dummyDate.minute = 0;
                    try {
                        dummyDate.normalize();
                    } catch(ex) {
                    }
                    startTime = getStringForDate(dummyDate);
                    if (noPrintTimesInMultiDay) {
                        startTime = "";
                    }
                } else {
                    startTime = getStringForDate(sDate);
                }
                if (eDate.compare(date) > 0) {
                    var dummyDate = eDate.clone();
                    dummyDate.hour = 23;
                    dummyDate.minute = 59;
                    try {
                        dummyDate.normalize();
                    } catch(ex) {
                    }
                    endTime = getStringForDate(dummyDate);
                    if (noPrintTimesInMultiDay) {
                        endTime = "";
                    }
                } else {
                    endTime = getStringForDate(eDate);
                }
                if (printEndTime) {
                    time = startTime + '-' + endTime;
                    if (time == "-") {
                        time = "";
                    }
                } else {
                    time = startTime;
                }
            } else {
                time = getStringForDate(sDate);
            }
        } else if (sDate) {
            time = getStringForDate(sDate);
        } else if (eDate) {
            time = getStringForDate(eDate);
        }

        try {
            var calColor = item.calendar.getProperty('color');
        } catch(ex) {
            var calMgr = Components.classes["@mozilla.org/calendar/manager;1"]
                                   .getService(Components.interfaces.calICalendarManager);
            var calColor = calMgr.getCalendarPref(item.calendar, 'color'); 
        }
        if (!calColor) {
            calColor = "#A8C2E1";
        }
        var pb2 = Components.classes["@mozilla.org/preferences-service;1"]
                            .getService(Components.interfaces.nsIPrefBranch2);
        var catColor;
        try {
            catColor = pb2.getCharPref("calendar.category.color."+item.getProperty("CATEGORIES").toLowerCase());
        } catch(ex) {}
        if (!catColor) {
            for each (var cat in item.getCategories({})) {
                try {
                    catColor = pb2.getCharPref("calendar.category.color." + cat.toLowerCase());
                    break; // take first matching
                } catch(ex) {}
            }
        }
        var borderType = 'solid';
        if (item.status && printStatus) {
            if (item.status == "TENTATIVE") {
                borderType = 'dotted';
            }
            if (item.status == "CANCELLED") {
                borderType = 'dashed';
            }
        }
        if (!oldcolour) {
            var style = 'font-size: '+fsize+'px; text-align: left;';
            var tStyle = 'font-size: '+fsizet+'px; text-align: left; color: black;';
            style += ' color: black; border: ' + borderType + ' ' + calColor + ' 2px; width: 100%; ';
            if (catColor) {
                style += 'border-right-width: 4px; border-right-color: ' + catColor + ';';
                style += 'border-top-color: ' + catColor + ';';
                catColor = 0;
            }
        } else {
            var style = 'font-size: '+fsize+'px; text-align: left; background-color: '+calColor + ';';
            style += ' color: ' + getContrastingTextColor(calColor) + ';';
            var tStyle = 'font-size: '+fsizet+'px; text-align: left; background-color: '+calColor + '; color: ' + getContrastingTextColor(calColor) + ';';
            if (catColor) {
                style += ' border: solid ' + catColor + ' 2px;';
                catColor = 0;
            }
        }
        if (printPriority && item.priority && !oldcolour) {
            var prioCol = '';
            if (item.priority > 5) {
                prioCol = "#CFC";
            }
            if (item.priority == 5) {
                prioCol = "#FFC";
            }
            if (item.priority < 5) {
                prioCol = "#FCC";
            }
            style += 'background-color: ' + prioCol + ';';
        }
        var title = item.title;
        var location = " ";
        var descr = " ";
        if (printLocation)
        {
            if (item.hasProperty('LOCATION')) {
                location = item.getProperty('LOCATION');
            }
        }
        if (printDescription)
        {
            if (item.hasProperty('DESCRIPTION')) {
                descr = item.getProperty('DESCRIPTION');
            }
        }
/*        if (printLocation) {
            if (item.hasProperty('LOCATION')) {
                title += " (" + item.getProperty('LOCATION') +")";
            }
        } */
        if (obeyPrivacy && item.privacy != "PUBLIC") {
            title = " ";
        }

//                       <td valign='top' width='100%' style={style}><span style={tStyle}>{time}</span> {title}</td>

        var item = <tr width='100%'>
                       <td valign='top' align='left' width='100%' style={style}><span style={tStyle}>{time}</span> {title}<div class='desc'>{descr}</div><div class='loc'>{location}</div></td>
                   </tr>;
        innerTable.appendChild(item);
    }
  }
    dayTd.appendChild(innerTable);
    if (smallDay > 1) {
        return innerTable;
    } else {
        return dayTd;
    }
};



FGcalMonthPrinter.prototype.makeHTMLWeek =
function FGmakeHTMLWeek(date, sortedList, targetMonth) {
    var weekRow = <tr height='85'/>;
    var obeyPrivacy = getPrefSafe("fgprinters.obeyprivacy", false);
    var printWeekNo = getPrefSafe("fgprinters.monthly.printweekno", false);
    const weekPrefix = "calendar.week.";
    var prefNames = ["d0sundaysoff", "d1mondaysoff", "d2tuesdaysoff",
                     "d3wednesdaysoff", "d4thursdaysoff", "d5fridaysoff", "d6saturdaysoff"];
    var defaults = [true, false, false, false, false, false, true];
    var daysOff = new Array();
    for (var i in prefNames) {
        if (getPrefSafe(weekPrefix+prefNames[i], defaults[i])) {
            daysOff.push(Number(i));
        }
    }

    var printAllDay = getPrefSafe("fgprinters.monthly.printallday", false);
    var printPriority = getPrefSafe("fgprinters.monthly.printpriority", true);
    var printStatus = getPrefSafe("fgprinters.monthly.printstatus", true);
    var fsize = getPrefSafe("fgprinters.monthly.fontsize", 10);
    var oldcolour = getPrefSafe("fgprinters.monthly.oldcolour", false);
    var printLocation = getPrefSafe("fgprinters.monthly.printlocation", false);
    var smallSatSun = getPrefSafe("fgprinters.monthly.smallsatsun", false);

    if (printWeekNo) {
        try {
            var weekFormatter = Components.classes["@mozilla.org/calendar/weektitle-service;1"]
                                          .getService(Components.interfaces.calIWeekTitleService);
        } catch(ex) {
            var weekFormatter = Components.classes["@mozilla.org/calendar/weekinfo-service;1"]
                                          .getService(Components.interfaces.calIWeekInfoService);
        }
        var weekno = weekFormatter.getWeekTitle(date);
        var weekNoTd = <td class='weekno' align='center'> {weekno} </td>;
        weekRow.appendChild(weekNoTd);
    }

    try {
		var defaultTimezone = calendarDefaultTimezone(); 
	} catch(ex) {
	    var defaultTimezone = null;
	}
    if (!smallSatSun) {
        for (var i = 0; i < 7; i++) {
            weekRow.appendChild(this.makeHTMLDay(date, sortedList, targetMonth, 0));
            date.day++;
            try {
                date.normalize();
            } catch(ex) {
            }
        }
    } else {
        for (var i = 0; i < 5; i++) {
            weekRow.appendChild(this.makeHTMLDay(date, sortedList, targetMonth, 0));
            date.day++;
            try {
                date.normalize();
            } catch(ex) {
            }
        }
        var sat = this.makeHTMLDay(date, sortedList, targetMonth, 1);
        date.day++;
        try {
            date.normalize();
        } catch(ex) {
        }
        var sun = this.makeHTMLDay(date, sortedList, targetMonth, 2);
        date.day++;
        try {
            date.normalize();
        } catch(ex) {
        }
        sat.appendChild(sun);
        weekRow.appendChild(sat);
//        var sunTr = <tr>{sun}</tr>;
//        var weekEnd = <tr><td height='100' width='100'><table border='0'><tr>{sat}</tr><tr>{sun}</tr></table></td></tr>;
//        weekRow = weekRow + sunTr; //.appendChild(weekEnd);
    }
    return weekRow;
};

FGcalMonthPrinter.prototype.createMonthTable =
function FGmonthPrint_doOtherMonth(aMonth) {
    var weekStart = getPrefSafe("calendar.week.start", 0);

	var monthName = calGetString("dateFormat", "month." + (aMonth.month +1)+ ".name") + " " + aMonth.year;

    var fsize = getPrefSafe("fgprinters.monthly.fontsizeom", 10);
    var tStyle = 'font-size: ' + fsize + 'px;';

    var monthTable = <table width='100%' rules='both' style={tStyle}><tr><td colspan='7' align='center'>{monthName}</td></tr></table>;
    var firstDate = aMonth.startOfMonth.startOfWeek.clone();
    firstDate.day += weekStart;
    try {
        firstDate.normalize();
    } catch(ex) {
    }
    if (aMonth.startOfMonth.weekday < weekStart) {
        // Go back one week to make sure we display this day
        firstDate.day -= 7;
        try {
            firstDate.normalize();
        } catch(ex) {
        }
    }

    var lastDate = aMonth.endOfMonth.endOfWeek.clone();
    if (aMonth.endOfMonth.weekday < weekStart) {
        // Go back one week so we don't display any extra days
        lastDate.day -= 7;
        try {
            lastDate.normalize();
        } catch(ex) {
        }
    }
    firstDate.isDate = true;
    lastDate.isDate = true;

    var printWeekDays = getPrefSafe("fgprinters.printweekdaysmm", false);
    if (printWeekDays) {
        const weekPrefix = "calendar.week.";
        var prefNames = ["d0sundaysoff", "d1mondaysoff", "d2tuesdaysoff",
                         "d3wednesdaysoff", "d4thursdaysoff", "d5fridaysoff", "d6saturdaysoff"];
        var defaults = [true, false, false, false, false, false, true];
        var daysOff = new Array();
        for (var i in prefNames) {
            if (getPrefSafe(weekPrefix+prefNames[i], defaults[i])) {
                daysOff.push(Number(i));
            }
        }
        var weekRow = <tr/>;
        for (var i = 0; i < 7; i++) {
            var myClass = 'day-box';
            var style = 'padding-left: 0.2em; padding-right: 0.2em; ';
            var dayName = calGetString("dateFormat", "day."+ (((i + weekStart) % 7)+1) + ".Mmm")[0];
            var day = <td align='center' valign='center' class={myClass}><div style={style}>{dayName}</div></td>
            weekRow.appendChild(day);
            try {
                date.normalize();
            } catch(ex) {
            }
        }
        monthTable.appendChild(weekRow);
    }

    var date = firstDate.clone();
    var itemListIndex = 0;
    while (date.compare(lastDate) != 1) {
        monthTable.appendChild(this.makeOtherMonthWeek(date, aMonth.month));
    }
    return monthTable;
};

FGcalMonthPrinter.prototype.makeOtherMonthWeek =
function FGmakeOtherMonthWeek(date, targetMonth) {
    var weekRow = <tr/>;
    var printOMDays = getPrefSafe("fgprinters.printomdaysmm", true);
    const weekPrefix = "calendar.week.";
    var prefNames = ["d0sundaysoff", "d1mondaysoff", "d2tuesdaysoff",
                     "d3wednesdaysoff", "d4thursdaysoff", "d5fridaysoff", "d6saturdaysoff"];
    var defaults = [true, false, false, false, false, false, true];
    var daysOff = new Array();
    for (var i in prefNames) {
        if (getPrefSafe(weekPrefix+prefNames[i], defaults[i])) {
            daysOff.push(Number(i));
        }
    }

    for (var i = 0; i < 7; i++) {
        var style = 'padding-left: 0.2em; padding-right: 0.2em;';
        var myClass = 'day-box';
        if (date.month != targetMonth) {
            myClass += ' out-of-month';
            if (!printOMDays) {
                style += 'color: transparent; ';
            }
        } else if (daysOff.some(function(a) { return a == date.weekday; })) {
            myClass += ' day-off';
        }
        var day = <td align='right' valign='center' class={myClass}><div style={style}>{date.day}</div></td>
        weekRow.appendChild(day);
        date.day++;
        try {
            date.normalize();
        } catch(ex) {
        }
    }
    return weekRow;
};


/**
 * Normal get*Pref calls will throw if the pref is undefined.  This function
 * will get a bool, int, or string pref.  If the pref is undefined, it will
 * return aDefault.
 *
 * @param aPrefName   the (full) name of preference to get
 * @param aDefault    (optional) the value to return if the pref is undefined
 */

function getPrefSafe(aPrefName, aDefault) {
    const nsIPrefBranch = Components.interfaces.nsIPrefBranch;
    const prefB = Components.classes["@mozilla.org/preferences-service;1"]
                            .getService(nsIPrefBranch);
    switch (prefB.getPrefType(aPrefName)) {
        case nsIPrefBranch.PREF_BOOL:
            return prefB.getBoolPref(aPrefName);
        case nsIPrefBranch.PREF_INT:
            return prefB.getIntPref(aPrefName);
        case nsIPrefBranch.PREF_STRING:
            return prefB.getCharPref(aPrefName);
        default: // includes nsIPrefBranch.PREF_INVALID
            return aDefault;
    }
}


/**
 * Gets the value of a string in a .properties file
 *
 * @param aBundleName  the name of the properties file.  It is assumed that the
 *                     file lives in chrome://calendar/locale/
 * @param aStringName the name of the string within the properties file
 */
function calGetString(aBundleName, aStringName) {
    var sbs = Components.classes["@mozilla.org/intl/stringbundle;1"]
                        .getService(Components.interfaces.nsIStringBundleService);
    var props = sbs.createBundle("chrome://calendar/locale/"+aBundleName+".properties");
    return props.GetStringFromName(aStringName);
}

function log(what) {
    var consoleSvc = Components.classes["@mozilla.org/consoleservice;1"].
                     getService(Components.interfaces.nsIConsoleService);
    consoleSvc.logStringMessage(what);    
}

/**
 * Function to get the (cached) best guess at a user's default timezone.  We'll
 * use the value of the calendar.timezone.local preference, if it exists.  If
 * not, we'll do our best guess.
 *
 * @return user's default timezone.
 */
function calendarDefaultTimezone() {
    if (calendarDefaultTimezone.mTz === undefined) {
        var prefTzid = getPrefSafe("calendar.timezone.local", null);
        var tzid = prefTzid;
        if (!tzid) {
            tzid = guessSystemTimezone();
        }
        calendarDefaultTimezone.mTz = getTimezoneService().getTimezone(tzid);
        ASSERT(calendarDefaultTimezone.mTz, "timezone not found: " + tzid);
        // Update prefs if necessary:
        if (calendarDefaultTimezone.mTz && calendarDefaultTimezone.mTz.tzid != prefTzid) {
            setPref("calendar.timezone.local", "CHAR", calendarDefaultTimezone.mTz.tzid);
        }
    }
    return calendarDefaultTimezone.mTz;
}

function getTimezoneService() {
    if (getTimezoneService.mObject === undefined) {
        getTimezoneService.mObject = Components.classes["@mozilla.org/calendar/timezone-service;1"]
                                               .getService(Components.interfaces.calITimezoneService);
    }
    return getTimezoneService.mObject;
}
