/* -*- Mode: javascript; tab-width: 20; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mozilla Calendar code.
 *
 * The Initial Developer of the Original Code is
 *   Joey Minta <jminta@gmail.com>
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * The output code for the time sheet was written by
 *   Ferdinand Grassmann <ferdinand@grassmann.info>
 *
 * Contributor(s):
 *   Matthew Willis <lilmatt@mozilla.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/**
 * Prints a time sheet view of a week of events
 */

function FGcalTimeSheetPrinter() {
    this.wrappedJSObject = this;
}

FGcalTimeSheetPrinter.prototype.QueryInterface =
function QueryInterface(aIID) {
    if (!aIID.equals(Components.interfaces.nsISupports) &&
        !aIID.equals(Components.interfaces.calIPrintFormatter)) {
        throw Components.results.NS_ERROR_NO_INTERFACE;
    }

    return this;
};

FGcalTimeSheetPrinter.prototype.getName =
function FGTimeSheetPrint_getName() {
    var sbs = Components.classes["@mozilla.org/intl/stringbundle;1"]
                    .getService(Components.interfaces.nsIStringBundleService);
    var props = sbs.createBundle("chrome://fgprinters/locale/fgprinters.properties");
    var retName = "FG " + props.GetStringFromName("timesheet");
    return retName;
};
FGcalTimeSheetPrinter.prototype.__defineGetter__("name", FGcalTimeSheetPrinter.prototype.getName);

FGcalTimeSheetPrinter.prototype.formatToHtml =
function FGTimeSheetPrint_format(aStream, aStart, aEnd, aCount, aItems, aTitle) {

    // Create the e4x framework of the HTML document
    var html = <html/>;
    html.appendChild(
            <head>
                <title>{aTitle}</title>
                <meta http-equiv='Content-Type' content='text/html; charset=UTF-8'/>
                <style type='text/css'/>
            </head>);

    var fsizeday = getPrefSafe("fgprinters.ts.fontsizedayname", 20);
    var fsizetitle = getPrefSafe("fgprinters.ts.fontsizetitle", 20);
    var dnbk = getPrefSafe("fgprinters.headcolour", "#f0f0f0");
    var wec = getPrefSafe("fgprinters.weekendcolour", "#FFFFEE");
    var sansserif = getPrefSafe("fgprinters.sansserif", true);
    var fsizeloc = getPrefSafe("fgprinters.ts.fontsizelocation", 14);
    var fsizedesc = getPrefSafe("fgprinters.ts.fontsizedescription", 14);
    var printLocationRight = getPrefSafe("fgprinters.ts.printlocationright", true);
    var rowsSameHeight = getPrefSafe("fgprinters.ts.rowssameheight", false);
    var printHourLines = getPrefSafe("fgprinters.ts.printhourlines", true);

    var printtitle = getPrefSafe("fgprinters.printtitle", false);
    var fsizeglobaltitle = getPrefSafe("fgprinters.fontsizetitle", 26);
    var fsizesubtitle = getPrefSafe("fgprinters.fontsizesubtitle", 20);
    var subTitle = getPrefSafe("fgprinters.subtitle", "");

    var seperateDays = getPrefSafe("fgprinters.ts.seperatedays", false);


    html.head.style = ".main-table { font-size: "+fsizetitle+"px; font-weight:bold; }\n";

    html.head.style += ".main-title { font-size: "+fsizeglobaltitle+"px; font-weight:bold; text-align: center; }\n";
    html.head.style += ".sub-title { font-size: "+fsizesubtitle+"px; font-weight:bold; text-align: center; }\n";

    html.head.style += ".day-name { border: 1px solid #000; background-color: "+dnbk+"; font-size: "+fsizeday+"px; font-weight: bold; }\n";
    if (sansserif) { html.head.style += "body, tr, td, table, p { font-family: Arial, sans-serif; }\n"; }
    html.head.style += ".weekend { background-color: "+wec+"; }\n";
    if (printLocationRight)
    {
        html.head.style += ".loc { font-size: "+fsizeloc+"px; text-align: right; }\n";
    } else {
        html.head.style += ".loc { font-size: "+fsizeloc+"px; text-align: left; }\n";
    }
    html.head.style += ".desc { font-size: "+fsizedesc+"px; text-align: left; }\n";
    
    var body = <body/>;

    if (printtitle && aTitle) {
        body.appendChild(<p class="main-title">{aTitle}</p>);
    }

    if (subTitle) {
        body.appendChild(<p class="sub-title">{subTitle}</p>);
    }

    // helper: returns the passed item's startDate, entryDate or dueDate, in
    //         that order. If the item doesn't have one of those dates, this
    //         doesn't return.
    function hasUsableDate(item) {
        return item.startDate || item.entryDate || item.dueDate;
    }

    // Clean out the item list so it only contains items we will want to
    // include in the printout.
    var filteredItems = aItems.filter(hasUsableDate);

    var calIEvent = Components.interfaces.calIEvent;
    var calITodo = Components.interfaces.calITodo;

    try {
        var defaultTimezone = calendarDefaultTimezone(); 
    } catch(ex) {
        var defaultTimezone = null;
    }

    function compareItems(a, b) {
        // Sort tasks before events
        if (a instanceof calIEvent && b instanceof calITodo) {
            return 1;
        }
        if (a instanceof calITodo && b instanceof calIEvent) {
            return -1;
        }

        if (a instanceof calIEvent) {
            if (a.startDate.isDate && b.startDate.isDate)
            {
                return a.startDate.compare(b.startDate);
            }
            if (a.startDate.isDate)
            {
                var tmpd = b.startDate.clone();
                tmpd.isDate = true;
                try {
                    tmpd.normalize();
                } catch(ex) {
                }
                if (defaultTimezone) {
                    try {
                        tmpd = tmpd.getInTimezone(defaultTimezone);
                    } catch(ex) {
                    }
                }
                if (a.startDate.compare(tmpd) == 0) return -1;
            }
            if (b.startDate.isDate)
            {
                var tmpd = a.startDate.clone();
                tmpd.isDate = true;
                try {
                    tmpd.normalize();
                } catch(ex) {
                }
                if (defaultTimezone) {
                    try {
                        tmpd = tmpd.getInTimezone(defaultTimezone);
                    } catch(ex) {
                    }
                }
                if (b.startDate.compare(tmpd) == 0) return 1;
            }
        }
        
        if (a instanceof calIEvent) {
            var startCompare = a.startDate.compare(b.startDate);
            if (startCompare != 0) {
                return startCompare;
            }
            return a.endDate.compare(b.endDate);
        }
        var aDate = a.entryDate || a.dueDate;
        var bDate = b.entryDate || b.dueDate;
        return aDate.compare(bDate);
    }
    var sortedList = filteredItems.sort(compareItems);

    var sbs = Components.classes["@mozilla.org/intl/stringbundle;1"]
                        .getService(Components.interfaces.nsIStringBundleService);
    var props = sbs.createBundle("chrome://calendar/locale/calendar.properties");

    try {
        var weekFormatter = Components.classes["@mozilla.org/calendar/weektitle-service;1"]
                                      .getService(Components.interfaces.calIWeekTitleService);
    } catch(ex) {
        var weekFormatter = Components.classes["@mozilla.org/calendar/weekinfo-service;1"]
                                      .getService(Components.interfaces.calIWeekInfoService);
    }

    // A place to cache the string for "All Day" events later if we need it.
    var mAllDayString = null;
    
    // Start at the beginning of the week that aStart is in, and loop until
    // we're at aEnd. In the loop we build the HTML table for each day, and
    // get the day's items using getDayTd().
    var start = aStart || sortedList[0].startDate || sortedList[0].entryDate ||
                sortList[0].dueDate;
    ASSERT(start, "can't find a good starting date to print");

    var lastItem = sortedList[sortedList.length-1];
    var end = aEnd || lastItem.startDate || lastItem.entryDate ||
               lastItem.dueDate;
    ASSERT(end, "can't find a good ending date to print");

    var date;

    if (!seperateDays) {
        date = start.startOfWeek;
        var startOfWeek = getPrefSafe("calendar.week.start", 0);
        date.day += startOfWeek;
        try {
            date.normalize();
        } catch(ex) {
        }
        // Make sure we didn't go too far ahead
        if (date.compare(start) == 1) {
            date.day -= 7;
            try {
                date.normalize();
            } catch(ex) {
            }
        }
    } else {
        date = start.clone();
    }
    
    const weekPrefix = "calendar.week.";
    var prefNames = ["d0sundaysoff", "d1mondaysoff", "d2tuesdaysoff",
                     "d3wednesdaysoff", "d4thursdaysoff", "d5fridaysoff", "d6saturdaysoff"];
    var defaults = [true, false, false, false, false, false, true];


    if (!seperateDays) {
        while(date.compare(end) == -1) {
            var weekno = weekFormatter.getWeekTitle(date);
            var weekTitle = props.formatStringFromName('WeekTitle', [weekno], 1);
            var mainWeek =  <table border='1' rules='cols' width='50px' class='main-table'>
                             <tr> 
                                 <td align='center' valign='bottom' colspan='99'>{weekTitle}</td>
                             </tr>
                         </table>;
    
            // Create the <td> for each day, and put it into an array.
            var dayTds = new Array();
            var dayTopTds = new Array();
            for (var i = 0; i < 7 ; i++) {
                var isOff = getPrefSafe(weekPrefix+prefNames[date.weekday], defaults[date.weekday]);
                dayTds[date.weekday] = this.getDayTd(date, sortedList, isOff);
                dayTopTds[date.weekday] = this.getAllDayTd(date, sortedList, isOff);
                date.day += 1;
                try {
                    date.normalize();
                } catch(ex) {
                }
            }

            var fsize = getPrefSafe("fgprinters.ts.fontsize", 12);
        
        // <td ...?
            var col = <td valign="top"/>
            var startDay = getPrefSafe("fgprinters.ts.starthour", 7);
            var displayHours = getPrefSafe("fgprinters.ts.displayhours", 15);
            var endDay = startDay + displayHours;
            if (endDay > 24) {
                endDay = 24;
                displayHours = endDay - startDay;
            }

            var posStyle = 'position: relative; top: 0px; color: #000; font-size: ' + fsize + 'px';
            var posStyleWhite = 'position: relative; top: 0px; color: #fff; font-size: ' + fsize + 'px; width: 3em;';
            var myTab = <table valign='top' border='0'><tr height='5px'></tr></table>;
            var myTd = <td valign="top"/>
            var myDiv = <div style={posStyle}> </div>;
            myDiv.appendChild(<div style={posStyleWhite}> </div>); 

            myTd.appendChild(myDiv);
            myTab.appendChild(myTd);
            col.appendChild(myTab);
        
            mainWeek.appendChild(col);
        
            // print allday-events

            for (var i = startOfWeek; i < 7 + startOfWeek; i++) {
                mainWeek.appendChild(dayTopTds[i % 7]);
            }
        
            // print horizontal lines and times

            if (printHourLines && (displayHours > 1))
            {
                var myLines = <tr height="0px"><td colspan="99"></td></tr>;
                var style='position: relative; top: 4px; font-size: ' + fsize + 'px; font-weight: normal;';
                var myTimes = <div style={style}> </div>;
                for (var horLines = 0; horLines < displayHours; horLines++)
                {
                    var myPosition = ((horLines * 500) / displayHours + horLines*0) - 2;
                    var style= 'position: absolute; top: ' + myPosition + 'px; width: 100%; border-top: solid #aaa 1px; ';
                    var oneLine = <div style={style}> </div>;
                    myTimes.appendChild(oneLine);
                }

                for (var times = startDay; times < endDay; times++)
                {
                    var timePos = ((times - startDay) * 500) / displayHours;
                    var style='position: absolute; top: ' + timePos + 'px; font-size: ' + fsize + 'px; font-weight: normal;';
                    myTimes.appendChild(<div style={style}>{times}:00</div>);
                } 
                myLines.appendChild(myTimes);
                mainWeek.appendChild(myLines); 
            }
        
            // print non-allday-events for the day
        
            var bottomTable = <tr><div style="position: relative; top: 0px;"></div></tr>;
            for (var i = startOfWeek; i < 7 + startOfWeek; i++) {
                bottomTable.appendChild(dayTds[i % 7]);
            }
            mainWeek.appendChild(bottomTable);

            body.appendChild(mainWeek);
            // Make sure each month gets put on its own page
            body.appendChild(<br style="page-break-after: always;"/>);
        }
    } else {
        while(date.compare(end) == -1) {
            var isOff = getPrefSafe(weekPrefix+prefNames[date.weekday], defaults[date.weekday]);
            var dayTd = this.getDayTd(date, sortedList, isOff);
            var dayTopTd = this.getAllDayTd(date, sortedList, isOff);
            date.day += 1;
            try {
                date.normalize();
            } catch(ex) {
            }

            var weekno = weekFormatter.getWeekTitle(date);
            var weekTitle = props.formatStringFromName('WeekTitle', [weekno], 1);

            var mainWeek =  <table border='1' rules='cols' width='100%' class='main-table'>
                             <tr> 
                                 <td align='center' valign='bottom' colspan='99'>{weekTitle}</td>
                             </tr>
                         </table>;


            var fsize = getPrefSafe("fgprinters.ts.fontsize", 12);
        
        // <td ...?
            var col = <td valign="top"/>
            var startDay = getPrefSafe("fgprinters.ts.starthour", 7);
            var displayHours = getPrefSafe("fgprinters.ts.displayhours", 15);
            var endDay = startDay + displayHours;
            if (endDay > 24) {
                endDay = 24;
                displayHours = endDay - startDay;
            }

            var posStyle = 'position: relative; top: 0px; color: #000; font-size: ' + fsize + 'px';
            var posStyleWhite = 'position: relative; top: 0px; color: #fff; font-size: ' + fsize + 'px; width: 3em;';
            var myTab = <table valign='top' border='0'><tr height='5px'></tr></table>;
            var myTd = <td valign="top"/>
            var myDiv = <div style={posStyle}> </div>;
            myDiv.appendChild(<div style={posStyleWhite}> </div>); 

            myTd.appendChild(myDiv);
            myTab.appendChild(myTd);
            col.appendChild(myTab);
        
            mainWeek.appendChild(col);
        
            // print allday-events
            mainWeek.appendChild(dayTopTd);
        
            // print horizontal lines and times

            if (printHourLines && (displayHours > 1))
            {
                var myLines = <tr height="0px"><td colspan="99"></td></tr>;
                var style='position: relative; top: 4px; font-size: ' + fsize + 'px; font-weight: normal;';
                var myTimes = <div style={style}> </div>;
                for (var horLines = 0; horLines < displayHours; horLines++)
                {
                    var myPosition = ((horLines * 500) / displayHours + horLines*0) - 2;
                    var style= 'position: absolute; top: ' + myPosition + 'px; width: 100%; border-top: solid #aaa 1px; ';
                    var oneLine = <div style={style}> </div>;
                    myTimes.appendChild(oneLine);
                }

                for (var times = startDay; times < endDay; times++)
                {
                    var timePos = ((times - startDay) * 500) / displayHours;
                    var style='position: absolute; top: ' + timePos + 'px; font-size: ' + fsize + 'px; font-weight: normal;';
                    myTimes.appendChild(<div style={style}>{times}:00</div>);
                } 
                myLines.appendChild(myTimes);
                mainWeek.appendChild(myLines); 
            }
        
            // print non-allday-events for the day
        
            var bottomTable = <tr><div style="position: relative; top: 0px;"></div></tr>;
            bottomTable.appendChild(dayTd);
            mainWeek.appendChild(bottomTable);

            body.appendChild(mainWeek);
            // Make sure each month gets put on its own page
            body.appendChild(<br style="page-break-after: always;"/>);
        }
    }
    html.appendChild(body);

    // Stream out the resulting HTML
    var convStream = Components.classes["@mozilla.org/intl/converter-output-stream;1"]
                               .getService(Components.interfaces.nsIConverterOutputStream);
    convStream.init(aStream, 'UTF-8', 0, 0x0000);
    convStream.writeString(html.toXMLString());
//    Components.utils.reportError(html.toXMLString());
};

function getContrastingTextColor(bgColor)
{
    var calcColor = bgColor.replace(/#/g, "");
    var red = parseInt(calcColor.substring(0, 2), 16);
    var green = parseInt(calcColor.substring(2, 4), 16);
    var blue = parseInt(calcColor.substring(4, 6), 16);

    // Calculate the L(ightness) value of the HSL color system.
    // L = (max(R, G, B) + min(R, G, B)) / 2
    var max = Math.max(Math.max(red, green), blue);
    var min = Math.min(Math.min(red, green), blue);
    var lightness = (max + min) / 2;

    // Consider all colors with less than 50% Lightness as dark colors
    // and use white as the foreground color; otherwise use black.
    // Actually we use a threshold a bit below 50%, so colors like
    // #FF0000, #00FF00 and #0000FF still get black text which looked
    // better when we tested this.
    if (lightness < 120) {
        return "white";
    }
    
    return "black";
}

FGcalTimeSheetPrinter.prototype.getAllDayTd =
/**
 * Given a calIDateTime and an array of items, this function creates an HTML
 * table containing the items, using the appropriate formatting and colours.
 */
function FGTimeSheetPrint_getAllDayTable(aDate, aItems, isOff) {
    // mainTd is the <td> element from the parent HTML table that will hold
    // the child HTML tables containing the date string and this day's items.
    var mainTd = <td border='0' width='100%' valign='top'/>
    var dateFormatter = Components.classes["@mozilla.org/calendar/datetime-formatter;1"]
                                  .getService(Components.interfaces.calIDateTimeFormatter);
    try {
        var defaultTimezone = calendarDefaultTimezone(); 
    } catch(ex) {
        var defaultTimezone = null;
    }
    try {
        var dateString = dateFormatter.dayName(aDate.getInTimezone(defaultTimezone).weekday) + " " + dateFormatter.formatDateShort(aDate.getInTimezone(defaultTimezone));
    } catch(ex) {
       var dateString = dateFormatter.dayName(aDate.weekday) + " " + dateFormatter.formatDateShort(aDate);
   }


    // Add the formatted date string (in its own child HTML table)
    mainTd.appendChild(
                     <table class='day-name' width='100%'>
                         <tr> 
                             <td align='center' valign='bottom'>{dateString}<div style='width: 130px;'> </div></td>
                         </tr>
                     </table>);

    var printAllDay = getPrefSafe("fgprinters.ts.printallday", false);
    var printLocation = getPrefSafe("fgprinters.ts.printlocation", true);
    var printDescription = getPrefSafe("fgprinters.ts.printdescription", true);
    var printPriority = getPrefSafe("fgprinters.ts.printpriority", true);
    var printStatus = getPrefSafe("fgprinters.ts.printstatus", true);
    var printEndTime = getPrefSafe("fgprinters.ts.printendtime", true);
    var obeyPrivacy = getPrefSafe("fgprinters.obeyprivacy", true);
    var startDay = getPrefSafe("fgprinters.ts.starthour", 7);
    var displayHours = getPrefSafe("fgprinters.ts.displayhours", 15);
    var noPrintTimesInMultiDay = getPrefSafe("fgprinters.ts.noprinttimesinmultidayevents", false);

    var endDay = startDay + displayHours;
    if (endDay > 24) {
        Components.utils.reportError("Too many hours to display (" + endDay + "), adjusting...");
        endDay = 24;
        displayHours = endDay - startDay;
    }

    var fsize = getPrefSafe("fgprinters.ts.fontsize", 12);
    var fsizet = getPrefSafe("fgprinters.ts.fontsizetime", 12);

    var oldcolour = getPrefSafe("fgprinters.ts.oldcolour", false);

        function getStringForDate1(aDate) {
            if (!aDate.isDate) {
                return dateFormatter.formatTime(aDate);
            }

            // We cache the string for "All Day" 'cause we're good like that.
            if (this.mAllDayString == null) {
                var sbs = Components.classes["@mozilla.org/intl/stringbundle;1"]
                                    .getService(Components.interfaces.nsIStringBundleService);
                var props = sbs.createBundle("chrome://calendar/locale/dateFormat.properties");
                this.mAllDayString = props.GetStringFromName("AllDay");
            }
            if (printAllDay) { return this.mAllDayString; } else { return ""; }
        }


        function isEventInRange1(item) {
            var sDate = item.startDate || item.entryDate || item.dueDate;
            var eDate = item.endDate || item.dueDate || item.entryDate;
            if ((sDate.compare(aDate) <= 0) && (eDate.compare(aDate) >= 0)) {
                return 1;
            }
        }

    var filteredEvents = aItems.filter(isEventInRange1);

    if (isOff) { var topTable = <table width='100%' valign='top' class="weekend" />; }
    else { var topTable = <table width='100%' valign='top'/>; }
    

    for each (var item in filteredEvents) {
        var sDate = item.startDate || item.entryDate || item.dueDate;
        var eDate = item.endDate || item.dueDate || item.entryDate;


        if (sDate) {
            try {
                sDate = sDate.getInTimezone(defaultTimezone);
            } catch(ex) {
            }
        }
        if (eDate) {
            try {
                eDate = eDate.getInTimezone(defaultTimezone);
            } catch(ex) {
            }
        }
        
        // End dates are exclusive. Adjust the eDate accordingly.
        if (sDate && sDate.isDate && eDate) {
            eDate = eDate.clone();
            eDate.day -= 1;
            try {
                eDate.normalize();
            } catch(ex) {
            }
        }

        // If the item has no end date, or if the item's end date is aDate or
        // is before aDate, skip to the next item.
        if (!eDate || (eDate.compare(aDate) < 0)) {
            continue;
        }

        // No start date or a start date that's after the date we want is bad.
        if (!sDate || (sDate.compare(aDate) > 0)) {
            break;
        }
        
        // only all day events!
        
        if ((sDate && !sDate.isDate) || (eDate && !eDate.isDate)) {
            continue;
        }

        var time;
        if (sDate && eDate) {
            if (!sDate.isDate) {
                var startTime = "";
                var endTime = "";
                if (sDate.compare(aDate) < 0) {
                    var dummyDate = sDate.clone();
                    dummyDate.hour = 0;
                    dummyDate.minute = 0;
                    try {
                        dummyDate.normalize();
                    } catch(ex) {
                    }
                    startTime = getStringForDate1(dummyDate);
                    if (noPrintTimesInMultiDay) {
                        startTime = "";
                    }
                } else {
                    startTime = getStringForDate1(sDate);
                }
                if (eDate.compare(aDate) > 0) {
                    var dummyDate = eDate.clone();
                    dummyDate.hour = 23;
                    dummyDate.minute = 59;
                    try {
                        dummyDate.normalize();
                    } catch(ex) {
                    }
                    endTime = getStringForDate1(dummyDate);
                    if (noPrintTimesInMultiDay) {
                        endTime = "";
                    }
                } else {
                    endTime = getStringForDate1(eDate);
                }
                if (printEndTime) {
                    time = startTime + '-' + endTime;
                    if (time == "-") {
                        time = "";
                    }
                } else {
                    time = startTime;
                }
            } else {
                time = getStringForDate1(sDate);
            }
        } else if (sDate) {
            time = getStringForDate1(sDate);
        } else if (eDate) {
            time = getStringForDate1(eDate);
        }

        // Get calendar and category colours and apply them to the item's
        // table cell.
        try {
            var calColor = item.calendar.getProperty('color');
        } catch(ex) {
            var calMgr = Components.classes["@mozilla.org/calendar/manager;1"]
                                .getService(Components.interfaces.calICalendarManager);
            var calColor = calMgr.getCalendarPref(item.calendar, 'color'); 
        }
        if (!calColor) {
            calColor = "#A8C2E1";
        }
        var pb2 = Components.classes["@mozilla.org/preferences-service;1"]
                            .getService(Components.interfaces.nsIPrefBranch2);
        var catColor;
        try {
            catColor = pb2.getCharPref("calendar.category.color."+item.getProperty("CATEGORIES").toLowerCase());
        } catch(ex) {}
        if (!catColor) {
            for each (var cat in item.getCategories({})) {
                try {
                    catColor = pb2.getCharPref("calendar.category.color." + cat.toLowerCase());
                    break; // take first matching
                } catch(ex) {}
            }
        }

        var borderType = 'solid';
        if (item.status && printStatus) {
            if (item.status == "TENTATIVE") {
                borderType = 'dotted';
            }
            if (item.status == "CANCELLED") {
                borderType = 'dashed';
            }
        }

        if (!oldcolour) {
            var style = 'font-size: '+fsize+'px; border: ' + borderType + ' ' + calColor + ' 4px; ';
            style += ' color: black; ';
            var tStyle = 'font-size: '+fsizet+'px; color: black;';
            if (catColor) {
                style += 'border-right-width: 8px; border-right-color: ' + catColor + ';';
                style += 'border-top-color: ' + catColor + ';';
                catColor = 0;
            }
        } else {
            var style = 'font-size: '+fsize+'px; font-weight: bold; background-color: ' + calColor + ';';
            style += ' color: ' + getContrastingTextColor(calColor) + ';';
            var tStyle = 'font-size: '+fsizet+'px; font-weight: bold; background-color: ' + calColor + '; color: ' + getContrastingTextColor(calColor) + ';';
            if (catColor) {
                style += ' border: solid ' + catColor + ' 2px;';
                catColor = 0;
            }
        }
        if (printPriority && item.priority && !oldcolour)
        {
            var prioCol = '';
            if (item.priority > 5) {
                prioCol = "#CFC";
            }
            if (item.priority == 5) {
                prioCol = "#FFC";
            }
            if (item.priority < 5) {
                prioCol = "#FCC";
            }
            style += 'background-color: ' + prioCol + ';';
        }

//    style += 'position: relative; top: ' + eStart + 'px; height: ' + eDuration + 'px;';
        var location = "";
        var descr = "";
        
        if (printLocation) {
            if (item.hasProperty('LOCATION')) {
                location = item.getProperty('LOCATION');
            }
        }
        if (printDescription) {
            if (item.hasProperty('DESCRIPTION')) {
                descr = item.getProperty('DESCRIPTION');
            }
        }
        var title = item.title;
        if (obeyPrivacy && item.privacy != "PUBLIC") {
            title = "";
            descr = "";
            location = "";
        }
        if (location && descr) {
            var item = <tr width='100%'>
                       <td valign='top' align='left' width='100%' style={style}>
               <span style={tStyle}>{time}</span> {title}
               <div class='desc'>{descr}</div>
               <div class='loc'>{location}</div></td>
                   </tr>; 
        }
        if (location && !descr) {
            var item = <tr width='100%'>
                       <td valign='top' align='left' width='100%' style={style}>
               <span style={tStyle}>{time}</span> {title}
               <div class='loc'>{location}</div></td>
                   </tr>; 
        }
        if (!location && descr) {
            var item = <tr width='100%'>
                       <td valign='top' align='left' width='100%' style={style}>
               <span style={tStyle}>{time}</span> {title}
               <div class='desc'>{descr}</div></td>
                   </tr>; 
        }
        if (!location && !descr) {
            var item = <tr width='100%'>
                       <td valign='top' align='left' width='100%' style={style}>
               <span style={tStyle}>{time}</span> {title}</td>
                   </tr>; 
        }
        topTable.appendChild(item);
    }
    topTable.appendChild(<tr height='100%'><td> </td></tr>);
    mainTd.appendChild(topTable);
    
    return mainTd;
};


FGcalTimeSheetPrinter.prototype.getDayTd =
/**
 * Given a calIDateTime and an array of items, this function creates an HTML
 * table containing the items, using the appropriate formatting and colours.
 */
function FGTimeSheetPrint_getDayTable(aDate, aItems, isOff) {
    // mainTd is the <td> element from the parent HTML table that will hold
    // the child HTML tables containing the date string and this day's items.
    var mainTd = <td border='0' width="100%" valign='top'/>
    var dateFormatter = Components.classes["@mozilla.org/calendar/datetime-formatter;1"]
                                  .getService(Components.interfaces.calIDateTimeFormatter);

    var printAllDay = getPrefSafe("fgprinters.ts.printallday", false);
    var printLocation = getPrefSafe("fgprinters.ts.printlocation", true);
    var printDescription = getPrefSafe("fgprinters.ts.printdescription", true);
    var printPriority = getPrefSafe("fgprinters.ts.printpriority", true);
    var printStatus = getPrefSafe("fgprinters.ts.printstatus", true);
    var printEndTime = getPrefSafe("fgprinters.ts.printendtime", true);
    var obeyPrivacy = getPrefSafe("fgprinters.obeyprivacy", true);
    var startDay = getPrefSafe("fgprinters.ts.starthour", 7);
    var displayHours = getPrefSafe("fgprinters.ts.displayhours", 15);
    var endDay = startDay + displayHours;
    var noPrintTimesInMultiDay = getPrefSafe("fgprinters.ts.noprinttimesinmultidayevents", false);


    if (endDay > 24) {
        Components.utils.reportError("Too many hours to display (" + endDay + "), adjusting...");
        endDay = 24;
        displayHours = endDay - startDay;
    }
    try {
        var defaultTimezone = calendarDefaultTimezone(); 
    } catch(ex) {
        var defaultTimezone = null;
    }
 
    var fsize = getPrefSafe("fgprinters.ts.fontsize", 12);
    var fsizet = getPrefSafe("fgprinters.ts.fontsizetime", 12);
    var oldcolour = getPrefSafe("fgprinters.ts.oldcolour", false);

        function getStringForDate(aDate) {
            if (!aDate.isDate) {
                return dateFormatter.formatTime(aDate);
            }

            // We cache the string for "All Day" 'cause we're good like that.
            if (this.mAllDayString == null) {
                var sbs = Components.classes["@mozilla.org/intl/stringbundle;1"]
                                    .getService(Components.interfaces.nsIStringBundleService);
                var props = sbs.createBundle("chrome://calendar/locale/dateFormat.properties");
                this.mAllDayString = props.GetStringFromName("AllDay");
            }
            if (printAllDay) { return this.mAllDayString; } else { return ""; }
        }


        function isEventInRange(item) {
            var sDate = item.startDate || item.entryDate || item.dueDate;
            var eDate = item.endDate || item.dueDate || item.entryDate;
            if ((sDate.compare(aDate) <= 0) && (eDate.compare(aDate) >= 0)) {
                return 1;
            }
        }

    var filteredEvents = aItems.filter(isEventInRange);

        function isNonAllDay(item) {
            var sDate = item.startDate || item.entryDate || item.dueDate;
            var eDate = item.endDate || item.dueDate || item.entryDate;
            if ((sDate && !sDate.isDate) && (eDate && !eDate.isDate)) {
                return 1;
            }
        }

    var filteredNonAllDayEvents = filteredEvents.filter(isNonAllDay);
    var alreadyPrinted = Array(filteredNonAllDayEvents.length);
    for (var eventIndex = 0; eventIndex < filteredNonAllDayEvents.length; eventIndex++) {
        alreadyPrinted[eventIndex] = 0;
    }

    if (isOff) { var innerTable = <table width='100%' valign='top' class="weekend" style='margin-top: -4px; margin-left: -1px; margin-right: -2px;'><tr height='500px'/></table>; }
    else { var innerTable = <table width='100%' valign='top' style='margin-top: -4px; margin-bottom: -4px; margin-left: -2px; margin-right: -2px;'><tr height='500px'/></table>; }

    if (!filteredNonAllDayEvents.length) {
        innerTable.appendChild(<td valign='top' align='left' width='100%' height='100%'/>);
    }

    for (var eventIndex = 0; eventIndex < filteredNonAllDayEvents.length; eventIndex++) {
        // don't bother if we already printed this one
        if (alreadyPrinted[eventIndex]) {
            continue;
        }
        alreadyPrinted[eventIndex] = 1;
        var item = filteredNonAllDayEvents[eventIndex];
        var sDate = item.startDate || item.entryDate || item.dueDate;
        var eDate = item.endDate || item.dueDate || item.entryDate;
        if (sDate) {
            try {
                sDate = sDate.getInTimezone(defaultTimezone);
            } catch(ex) {
            }
        } 
        if (eDate) {
            try {
                eDate = eDate.getInTimezone(defaultTimezone);
            } catch(ex) {
            } 
        }

        // End dates are exclusive. Adjust the eDate accordingly.
        if (sDate && sDate.isDate && eDate) {
            eDate = eDate.clone();
            eDate.day -= 1;
            try {
                eDate.normalize();
            } catch(ex) {
            }
        }

        // If the item has no end date, or if the item's end date is aDate or
        // is before aDate, skip to the next item.
        if (!eDate || (eDate.compare(aDate) < 0)) {
            continue;
        }

        // No start date or a start date that's after the date we want is bad.
        if (!sDate || (sDate.compare(aDate) > 0)) {
            break;
        }

        var printItem = <td valign='top' align='left' width='100%'/>;
        function addItemToTd(event, prev)
        {
            var sDate = event.startDate || event.entryDate || event.dueDate;
            var eDate = event.endDate || event.dueDate || event.entryDate;
            if (sDate) {
                try {
                    sDate = sDate.getInTimezone(defaultTimezone);
                } catch(ex) {
                }
            }
            if (eDate) {
                try {
                    eDate = eDate.getInTimezone(defaultTimezone);
                } catch(ex) {
                }
            }
            var subPrev = 0;
    
            // End dates are exclusive. Adjust the eDate accordingly.
            if (sDate && sDate.isDate && eDate) {
                eDate = eDate.clone();
                eDate.day -= 1;
                try {
                    eDate.normalize();
                } catch(ex) {
                }
            }
            var time;
            if (sDate && eDate) {
                if (!sDate.isDate) {
                    var startTime = "";
                    var endTime = "";
                    if (sDate.compare(aDate) < 0) {
                        var dummyDate = sDate.clone();
                        dummyDate.hour = 0;
                        dummyDate.minute = 0;
                        try {
                            dummyDate.normalize();
                        } catch(ex) {
                        }
                        startTime = getStringForDate(dummyDate);
                        if (noPrintTimesInMultiDay) {
                            startTime = "";
                        }
                    } else {
                        startTime = getStringForDate(sDate);
                    }
                    if (eDate.compare(aDate) > 0) {
                        var dummyDate = eDate.clone();
                        dummyDate.hour = 23;
                        dummyDate.minute = 59;
                        try {
                            dummyDate.normalize();
                        } catch(ex) {
                        }
                        endTime = getStringForDate(dummyDate);
                        if (noPrintTimesInMultiDay) {
                            endTime = "";
                        }
                    } else {
                        endTime = getStringForDate(eDate);
                    }
                    if (printEndTime) {
                        time = startTime + '-' + endTime;
                        if (time == "-") {
                            time = "";
                        }
                    } else {
                        time = startTime;
                    }
                } else {
                    time = getStringForDate(sDate);
                }
            } else if (sDate) {
                time = getStringForDate(sDate);
            } else if (eDate) {
                time = getStringForDate(eDate);
            }
    
            // Get calendar and category colours and apply them to the item's
            // table cell.
            try {
                var calColor = item.calendar.getProperty('color');
            } catch(ex) {
                var calMgr = Components.classes["@mozilla.org/calendar/manager;1"]
                                    .getService(Components.interfaces.calICalendarManager);
                var calColor = calMgr.getCalendarPref(item.calendar, 'color'); 
            }
            if (!calColor) {
                calColor = "#A8C2E1";
            }
            var pb2 = Components.classes["@mozilla.org/preferences-service;1"]
                                .getService(Components.interfaces.nsIPrefBranch2);
            var catColor;
            try {
                catColor = pb2.getCharPref("calendar.category.color."+event.getProperty("CATEGORIES").toLowerCase());
            } catch(ex) {}
            if (!catColor) {
                for each (var cat in item.getCategories({})) {
                    try {
                        catColor = pb2.getCharPref("calendar.category.color." + cat.toLowerCase());
                        break; // take first matching
                    } catch(ex) {}
                }
            }
            
            var begin = sDate.subtractDate(aDate).inSeconds / 60;
            var end = eDate.subtractDate(aDate).inSeconds / 60;
            if (begin < startDay * 60) { begin = startDay * 60; }
            if (begin > endDay * 60) { begin = endDay * 60; }
            if (end > endDay * 60) { end = endDay * 60; }
            if (end < startDay * 60) { end = startDay * 60; }
            begin -= startDay * 60;
            end -= startDay * 60;
            if (end < begin) {
                end = begin;
            }
            var eDuration = Math.round((end - begin) * 500 / (displayHours * 60)) - 4; // double height of border
            if (eDuration < 0) {
                eDuration = 2;
            }
            var borderType = 'solid';
            if (item.status && printStatus) {
                if (item.status == "TENTATIVE") {
                    borderType = 'dotted';
                }
                if (item.status == "CANCELLED") {
                    borderType = 'dashed';
                }
            }
            
            var eStart = Math.round(begin * 500 / (displayHours * 60)) - prev;
            if (!oldcolour) {
                var style = 'font-size: '+fsize+'px; ';
                var tStyle = 'font-size: '+fsizet+'px; color: black;';
                if (eDuration) {
                    style += 'border: ' + borderType + ' ' + calColor + ' 2px; ';
                }
                style += ' color: black; ';
                if (catColor && eDuration) {
                    style += 'border-right-width: 4px; border-right-color: ' + catColor + ';';
                    style += 'border-top-color: ' + catColor + ';';
                    catColor = 0;
                }
            } else {
                var style = 'font-size: '+fsize+'px; font-weight: bold; background-color: ' + calColor + ';';
                style += ' color: ' + getContrastingTextColor(calColor) + ';';
                var tStyle = 'font-size: '+fsizet+'px; font-weight: bold; background-color: ' + calColor + '; color: ' + getContrastingTextColor(calColor) + ';';
                if (catColor) {
                    style += ' border: solid ' + catColor + ' 2px;';
                    catColor = 0;
                }
            }
            style += 'margin-top: ' + eStart + 'px; height: ' + eDuration + 'px; margin-left: -2px; margin-right: -2px; ';
            if (printPriority && item.priority && !oldcolour)
            {
                var prioCol = '';
                if (item.priority > 5) {
                    prioCol = "#CFC";
                }
                if (item.priority == 5) {
                    prioCol = "#FFC";
                }
                if (item.priority < 5) {
                    prioCol = "#FCC";
                }
                style += 'background-color: ' + prioCol + ';';
            }
            var location = "";
            var descr = "";
            
            if (printLocation) {
                if (event.hasProperty('LOCATION')) {
                    location = event.getProperty('LOCATION');
                }
            }
            if (printDescription) {
                if (event.hasProperty('DESCRIPTION')) {
                    descr = event.getProperty('DESCRIPTION');
                }
            }
            var title = item.title;
            if (obeyPrivacy && item.privacy != "PUBLIC") {
                title = "";
                descr = "";
                location = "";
            }
            if (location && descr) {
                printItem.appendChild(
                   <div style={style}><span style={tStyle}>{time}</span> {title}
                   <div class='desc'>{descr}</div>
                   <div class='loc'>{location}</div></div>); 
            }
            if (location && !descr) {
                printItem.appendChild(
                   <div style={style}><span style={tStyle}>{time}</span> {title}
                   <div class='loc'>{location}</div></div>); 
            }
            if (!location && descr) {
                printItem.appendChild(
                   <div style={style}><span style={tStyle}>{time}</span> {title}
                   <div class='desc'>{descr}</div></div>); 
            }
            if (!location && !descr) {
                printItem.appendChild(
                   <div style={style}><span style={tStyle}>{time}</span> {title}</div>); 
            }
            return eStart+eDuration+prev +4; // width of bottom border
        }
        
        subPrev = addItemToTd(item, 0);
        var compareDate = eDate.clone();
        
        for (var newIndex = eventIndex; newIndex < filteredNonAllDayEvents.length; newIndex++) {
            if (alreadyPrinted[newIndex]) {
                continue;
            }
            var item = filteredNonAllDayEvents[newIndex];
            var sDate = item.startDate || item.entryDate || item.dueDate;
            var eDate = item.endDate || item.dueDate || item.entryDate;
            if (sDate) {
                try {
                    sDate = sDate.getInTimezone(defaultTimezone);
                } catch(ex) {
                }
            }
            if (eDate) {
                try {
                    eDate = eDate.getInTimezone(defaultTimezone);
                } catch(ex) {
                }
            }

            // End dates are exclusive. Adjust the eDate accordingly.
            if (sDate && sDate.isDate && eDate) {
                eDate = eDate.clone();
                eDate.day -= 1;
                try {
                    eDate.normalize();
                } catch(ex) {
                }
            }
            if (compareDate.compare(sDate) <= 0) {
                alreadyPrinted[newIndex] = 1;
                subPrev = addItemToTd(item, subPrev);
//                Components.utils.reportError(prevDur);
                compareDate = eDate.clone();
            }
        }
        innerTable.appendChild(printItem);
    }
    mainTd.appendChild(innerTable);
    return mainTd;
};

/**
 * Normal get*Pref calls will throw if the pref is undefined.  This function
 * will get a bool, int, or string pref.  If the pref is undefined, it will
 * return aDefault.
 *
 * @param aPrefName   the (full) name of preference to get
 * @param aDefault    (optional) the value to return if the pref is undefined
 */
function getPrefSafe(aPrefName, aDefault) {
    const nsIPrefBranch = Components.interfaces.nsIPrefBranch;
    const prefB = Components.classes["@mozilla.org/preferences-service;1"]
                            .getService(nsIPrefBranch);
    switch (prefB.getPrefType(aPrefName)) {
        case nsIPrefBranch.PREF_BOOL:
            return prefB.getBoolPref(aPrefName);
        case nsIPrefBranch.PREF_INT:
            return prefB.getIntPref(aPrefName);
        case nsIPrefBranch.PREF_STRING:
            return prefB.getCharPref(aPrefName);
        default: // includes nsIPrefBranch.PREF_INVALID
            return aDefault;
    }
}

/**
 * Logs a message and the current js-stack, if aCondition fails
 *
 * @param aCondition  the condition to test for
 * @param aMessage    the message to report in the case the assert fails
 * @param aCritical   if true, throw an error to stop current code execution
 *                    if false, code flow will continue
 */
function ASSERT(aCondition, aMessage, aCritical) {
    if (aCondition) {
        return;
    }

    var string = "Assert failed: " + aMessage + '\n' + STACK();
    if (aCritical) {
        throw new Error(string);
    } else {
        Components.utils.reportError(string);
    }
}


/**
 * Function to get the (cached) best guess at a user's default timezone.  We'll
 * use the value of the calendar.timezone.local preference, if it exists.  If
 * not, we'll do our best guess.
 *
 * @return user's default timezone.
 */
function calendarDefaultTimezone() {
    if (calendarDefaultTimezone.mTz === undefined) {
        var prefTzid = getPrefSafe("calendar.timezone.local", null);
        var tzid = prefTzid;
        if (!tzid) {
            tzid = guessSystemTimezone();
        }
        calendarDefaultTimezone.mTz = getTimezoneService().getTimezone(tzid);
        ASSERT(calendarDefaultTimezone.mTz, "timezone not found: " + tzid);
        // Update prefs if necessary:
        if (calendarDefaultTimezone.mTz && calendarDefaultTimezone.mTz.tzid != prefTzid) {
            setPref("calendar.timezone.local", "CHAR", calendarDefaultTimezone.mTz.tzid);
        }
    }
    return calendarDefaultTimezone.mTz;
}

function getTimezoneService() {
    if (getTimezoneService.mObject === undefined) {
        getTimezoneService.mObject = Components.classes["@mozilla.org/calendar/timezone-service;1"]
                                               .getService(Components.interfaces.calITimezoneService);
    }
    return getTimezoneService.mObject;
}
